;;;-*- Mode: Lisp; Encoding: (win32:code-page :id 1251) -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Basic variables, classes and generics
;;;
;;; General principles and naming convention: 
;;; (1) Every (even external) method with a name containing on-, confirm-, destroy-,
;;;     invalidate-, or %  must be invoked within 
;;;	(capi:execute-with-interface grid ...) or (capi:apply-in-pane-process grid ...)
;;; (2) When some external sym is provided with capi:execute-with-interface,
;;;	the correspoinding internal starts with %, e.g. %sym.
;;; (3) Every component draws the border or gridline to the right and bottom of itself.

(in-package :ystok.grid)

(defvar *default-press-return* nil
  "Press Enter key reaction: :down, :action or, otherwise, which interpreted like Tab.")
(defvar *confirm-deletions* t "Ask for confirmation of interactive deletion")

(defparameter *records-increment* 20 "Extension passed to vector-push-extend")
#+deprecated
(defparameter *use-os-cursors* t) ; Use Win32 |LoadCursorFromFile| for cursor images

;; Outer area background and text colors
(defparameter *workspace-background* #-(and lispworks4 win32) :color_appworkspace
                                     #+(and lispworks4 win32) 'win32:color_appworkspace)
(defparameter *workspace-foreground* :gray)

;;; Record navigation toolbar stuff
(defparameter *navigation-font* (gp:gf nil "arial" :normal :roman 6))

#|(gp:register-image-translation 'navigation-images
    #.(gp:read-external-image (merge-pathnames "images/recnavi.bmp"
                                        (or *compile-file-truename* *load-truename*))
                              :transparent-color-index 1)) ; =7 for 16-color bitmap
(defparameter *navigation-image-set*
  (capi:make-general-image-set :id #.(gp:read-external-image	;'navigation-images
                                      (lw:current-pathname "images/recnavi.bmp")
                                      :transparent-color-index 1)
                               :image-count 5 :width 10 :height 7))|#

(defparameter *default-column-width* 75)

;; Area around row or column border where cursor changes its appearence
(defparameter *header-resize-cursor-tolerance* 4)

(defclass basic-column ()
 (;; Identifying symbol or string, coinciding with SQL column name
  (name :accessor column-name :initarg :name)
  ;; Apparent columns are enumerated as 0,1,...; concealed - -1,-2,...
  (number :accessor column-number :initarg :number :type fixnum)
  (nullable :initarg :nullable :accessor column-nullable :initform t)
  ;; data -> cell-value, default is an integer index within the sequence of columns
  ;; Grid internals always wrap the call inside the template:
  ;;   (ignore-errors (call-data-reader grid (record-data record) column))
  ;; So make sure the function returns NIL as the second value on success. Otherwise,
  ;; that value would be mistreated as a condition and printed instead of the first value.
  (data-reader :accessor column-data-reader :initarg :data-reader)	;:initform nil
  ;; cell-value -> data, integer means index; nil if column is READ-ONLY
  ;; Serves as before_update: check validation rule or constraint
  ;; Either function with signature: (new-value data-object)
  ;; or (default) integer subscripting an element of the sequence.
  (data-writer :accessor column-data-writer :initarg :data-writer))
 (:documentation "Abstract grid column superclass"))

(defmethod print-object ((self basic-column) stream)
  (print-unreadable-object (self stream :type t)
    (format stream "~s nr ~d" (column-name self) (column-number self))))

(defclass column (basic-column)
 ((title :reader %column-title :initarg :title	; alternative title displayed on header
         :initform nil)				; if null, princ column-name instead
  (width :reader column-width :initarg :width			; in pixels
         :type fixnum
         :initform *default-column-width*)
  ;; Header horizontal alignment
  (header-align :accessor column-header-align
                :initarg :header-align :initarg :title-adjust
                :initform :center)
  ;; Can receive focus and be modified
  (enabled :reader enabled :initarg :enabled :initform t)
  (selected :accessor column-selected :initform nil)		; inside selection
  (help-key :reader help-key :initarg :help-key :initform nil)	; if T, use column name
  (graphics-args :accessor column-graphics-args			; plist :foreground
	:initarg :graphics-args :initform nil)			;	:background
  (sort-asc :reader column-sort-asc :initarg :sort-asc		; Signature: (x y)
	:initform nil)
  (sort-desc :reader column-sort-desc :initarg :sort-desc	; Signature: (x y)
	:initform nil)
  ;(sorted :accessor column-sorted :initform nil)      		; ::= :asc | :desc | nil
    ;; Cell enter callback. Signature: (grid record column)
    ;(enter-callback :initarg :enter-callback :initform nil)
    ;; Cell exit callback. Signature: (grid record column)
    ;(exit-callback :initarg :exit-callback :initform nil)
  ;; Double clicking on a cell. Signature: (grid record column &optional ratio)
  (cell-action-callback :initarg :action-callback :initarg :cell-action-callback
	:initform nil)
  ;; If header click is specified, this overwrite usual column selection behavior,
  ;; e.g. reverse-sorting. Signature: (grid column)
  (header-click-callback :initarg :header-click-callback :initform nil)
  ;; Header title text, perhaps wrapped
  (%cache% :accessor column-%cache% :initform *unbound-value*))
 (:documentation "Visible grid column"))

(defun column-title (column)
  (or (%column-title column) (column-name column)))

(defmethod ywi:object-title ((column basic-column))
  (column-title column))

(defmethod initialize-instance :after ((self column) &key (enabled t) graphics-args)
  (or enabled ;(enabled self)
      graphics-args
      (setf (column-graphics-args self)
             (list :foreground #-(and lispworks4 win32) :color_appworkspace
                               #+(and lispworks4 win32) 'win32:color_appworkspace))))

(defparameter *column-keyword-type-alist*
  `((:text	. text-column)
    (:display	. text-display-column)
    (:check	. check-column)))

(defun make-column (name number initargs &key (default-ground-col number)
                                              reader writer)
  ;; Args: initargs Come from parse-column-descriptions, not from :default-initargs!
  ;;       reader, writer  Default reader and writers
  ;; Values: <1> column object
  ;;	     <2> groundp
  (let* ((type (getf initargs :type :text))
         ;(data-writer (getf initargs :data-writer *unbound-value*))
  	 ;; by default, column with explicit or implicit data writer is not ghost
         ;(ground (getf initargs :ground data-writer))
         (column (apply #'make-instance (cdr-assoq3 type *column-keyword-type-alist* type)
                        :name name :number number
                        :allow-other-keys t
                        (remove-properties initargs '(:type :ground)))))
    (unless (slot-boundp column 'data-reader)
      (setf (column-data-reader column) (or reader default-ground-col)))
    (unless (slot-boundp column 'data-writer)
      (setf (column-data-writer column) (or writer default-ground-col)))
    (values column (getf initargs :ground (column-data-writer column)))))
            ;(append (remove-properties initargs '(:type :ground))
            ;        (when ground
            ;          (list :data-reader  (or reader default-ground-col)))
            ;        (when (and ground (eq data-writer *unbound-value*))
            ;          (list :data-writer (or writer default-ground-col))))
     ;ground)))

(defstruct record       ; Possible subclass of capi:item
  row-number            ; position where display in grid, nil if deleted
  data                  ; data object providing a (temporal) site for ground cell values
                        ; - list, vector, or whatever; nil for auto record
  selected              ; inside selection :accessor capi:item-selected
  plist                 ; Property indicators: :changed, :new, :deleted, :locked etc.
			; Simple vectors of cache objects for apparent columns
  (%cache% nil)         ; element ::= "str" | ("str" ...) | NIL | :true | :false
  (%graphics-cache% *unbound-value*))	; nil or simple vector of plists

(defmethod print-object ((self record) stream)
  (print-unreadable-object (self stream :type t)
    (format stream "~D~:[~; selected~]~{ ~S~}"
            (record-row-number self) (record-selected self) (record-plist self))))

(defmethod changed ((record record))
  (getf (record-plist record) :changed))
(defmethod (setf changed) (value (record record))
  (declare (optimize (debug 0) (speed 3) (safety 0)))
  (setf (getf (record-plist record) :changed) value))


;;; Dynamic scope established by call-data-reader, call-data-writer etc.
;;; Needed, e.g. when data-reader embeds call to the cv macro

(defvar *column*)

;; Incomplete list of binding:
;;  %fill-auto-record (calling addition-maker),
;;  #<method operate-on-grid (T ygrid (eql :append))>
(defvar *grid*)

;; Bound for cv and other exotic application (test-function etc.) Complete list:
;;  call-data-reader, write-cell-text, sort-records, compile-search-predicate,
;;  ywi:load-editing-pane, ywi:read-displayer-data, %write-cell-data-from-pane
(defvar *record-data*)

(defun call-data-reader (grid data column) ;(grid column data)
 ;;; Args: grid	Is insignificant unless data reader does refer to 
  ;;		other columns by means of the cv macro.
  (if data
      (let ((data-reader (column-data-reader column)))
        (if (integerp data-reader)
            (elt data data-reader)
            (let ((*grid* grid)
                  (*record-data* data)
                  (*column* column))
              (funcall data-reader data))))
      nil))  ; new auto-record at the bottom; return *unbound-value* instead?

(defun call-data-writer (value grid data column)
  (let ((data-writer (column-data-writer column)))
    (if (integerp data-writer)
        (setf (elt data data-writer) value)
        (let ((*grid* grid)
              (*column* column))
          (funcall data-writer value data)))))

(defmethod column-data-writer ((data-writer function))
 ;;; Allows (something . data-writer) within the changed alist for the sake of revert.
  data-writer)

(defmacro cv (column-designator)
 "Returns the value of the column designated by the argument
 within the dynamic scope of *grid*, *column, and *record-data*"
  (let* (data-reader			; cache data reader in the closure lexical, so
         column
         (function #'(lambda () ;(data)	; we don't need to extract it from the column
                      (unless data-reader
                        (setq column (find-column *grid* column-designator :errorp t)
                              data-reader (column-data-reader column)))
                      (when *record-data*
                        (if (integerp data-reader)
                            (elt *record-data* data-reader)
                            (let ((*column* column))
                              (funcall data-reader *record-data*)))))))
    `(funcall ,function))) ; *record-data*


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  GRID-COMPONENT  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass grid-component ()
 ((%font-height% :reader %font-height%)               	; cache values
  (%font-ascent% :reader %font-ascent%)
   ;(font-descent :initarg :font-descent :accessor font-descent)
  (border :reader border :initform t))			; True = must draw outer borders
 ;(%inversed-graphics-args% :initform nil)
 #+lispworks6.1
 (:default-initargs :drawing-mode :compatible))

;(defgeneric component-post-menu (grid-component x y &optional gesture)
; (:method ((grid-component grid-component) x y &optional gesture)
;  (declare (ignore grid-component x y gesture))))

(defgeneric component-on-create (grid-component)
 (:method ((grid-component grid-component))
  (unless (slot-boundp grid-component '%font-height%)	; the only way to determine
   (setf (slot-value grid-component '%font-height%) (gp:get-font-height grid-component)
         (slot-value grid-component '%font-ascent%) (gp:get-font-ascent grid-component)))))

;; Standard mouse and keyboard gestures
(defparameter *component-input-model*
 '((:post-menu		pane-post-menu)
   (:control-post-menu	pane-post-menu :control)
   (:keyboard-post-menu	pane-post-menu)

   ((:gesture-spec :prior :control)
    CAPI::CALL-FUNCTION-WITH-ARG-IGNORING-EXTRA-ARGS capi::%change-tab-page -1)
   ((:gesture-spec :next :control)
    CAPI::CALL-FUNCTION-WITH-ARG-IGNORING-EXTRA-ARGS capi::%change-tab-page 1)))

(defparameter *pinboard-input-model*
  (append
   '(((:button-1 :press)		pinboard-on-select)
     ((:button-1 :press :control)	pinboard-on-select :control)
     ((:button-1 :press :shift)		pinboard-on-select :shift)
     ((:button-1 :second-press)		pinboard-on-select :second-press)
     ;; Key gestures
     ((:gesture-spec :left)		pinboard-on-move)
     ((:gesture-spec :right)		pinboard-on-move)
     ((:gesture-spec :left  :control)	pinboard-on-move)
     ((:gesture-spec :right :control)	pinboard-on-move)
     ((:gesture-spec :prior :meta)	pinboard-on-move :page-left)
     ((:gesture-spec :next  :meta)	pinboard-on-move :page-right)
     ((:gesture-spec :home  :control)	pinboard-on-move)		; First column
     ((:gesture-spec :end   :control)	pinboard-on-move)		; Last column

     ((:gesture-spec :down)		pinboard-on-move)
     ((:gesture-spec :up)			pinboard-on-move)
     ((:gesture-spec :prior)		pinboard-on-move :page-up)	; PgUp
     ((:gesture-spec :next)		pinboard-on-move :page-down)	; PgDn
     ((:gesture-spec :up    :control)	pinboard-on-move :first)	; First record
     ((:gesture-spec :down  :control)	pinboard-on-move :last)		; Last record
     ((:gesture-spec #\Tab)		pinboard-on-move)
     ((:gesture-spec #\Tab  :shift)	pinboard-on-move))
   *component-input-model*
   '((:gesture-spec record-board-on-gesture))))				; character gestures

;;; Standard addition makers
;;; Args: None, but *grid* is bound during call.
;;; Values:
;;;  <1> data,
;;;  <2> list of modified columns/designators,
;;;  <3> slot names (defaults to T) pushed into :new; these slots
;;; 	 - have no corresponding column,
;;;	 - used by sql-datasheets only,
;;;	 - are similar to concealed columns of odbc-grid.
(defun list-addition-maker ()
  (make-list (ground-ncols *grid*)))
(defun vector-addition-maker ()
  #+lispwork4 (SEQ::MAKE-SIMPLE-VECTOR (ground-ncols *grid*))
  #-lispwork4 (make-array (ground-ncols *grid*)))

(defclass record-board (grid-component ywi:editable-pinboard)
 ((%last-scroll-dimension% :initform nil)
  (%last-scroll-operation% :initform nil)
  (button1-press-internal-time :accessor button1-press-internal-time :initform 0)
  ;; the buffer to be set in #<method ywi:make-editing-pane ((eql 'capi:editor-pane)...)>
  (buffer :reader buffer :initform nil)
  (%mono-image% :reader %mono-image% :initform nil))	; gp:image to be drawn on buttons
 (:default-initargs 
  :input-model *pinboard-input-model*
  :visible-min-width 50 :visible-min-height 1
  :resize-callback 'pinboard-on-resize
  :pane-can-scroll t
  :vertical-scroll t  
  :scroll-initial-y 0				; initial slug position
  :scroll-start-y 0  :scroll-height 0		; scroll range
  :scroll-vertical-step-size 10			; scroll increments
  :scroll-vertical-page-size 50
  :scroll-callback 'pinboard-on-scroll
  :focus-callback 'pinboard-on-focus
  :edit-on-enter t  :save-on-exit t		; from editable-pinboard
))

(defmethod ywi:object-title ((pinboard record-board))
  (when-let (column (current-column (capi:element-interface pinboard)))
    (ywi:object-title column)))

(defun record-number-pane-on-edit (pane #-lispworks4 #-lispworks4 &optional operation)
  (message-area-format pane
                       #-lispworks4
                       (if (eq operation :start)
                           #L"Enter record number and press ENTER to go there."
                           +null-string+)
                       #+lispworks4
                       #L"Enter record number and press ENTER to go there."))

(yl:keep-symbol-names 'record-number-pane)

(defparameter *navigation-button-alist*
  '((:first  :image 0)					; recnavi.bmp
    (:up     :image 1)
    (:down   :image 2)
    (:last   :image 3)
    (:new    :image 4)
    (save-record   :image 0 :tooltip #L"Save record")		; recoper.bmp
    (delete-record :image 1 :tooltip #L"Delete record")))

(defmacro navigation-button-image (data)
  `(getf (cdr-assoq ,data *navigation-button-alist*) :image))

(defclass navigation-bar (capi:toolbar)
  ;; CAPI objects initialized later
 ((records-title :reader records-title :initform nil)	; ::= empty, "Record" or the like
  (record-number-pane :reader record-number-pane :initform nil)
  (records-count-title :reader records-count-title :initform nil))
 (:default-initargs
  :items `(:first :up
           ,(make-instance 'capi:text-input-pane
                           :name 'record-number-pane
                           :callback 'record-number-callback
                           :callback-type :interface-data
 #+lispworks4 #+lispworks4 :change-callback 'record-number-pane-on-edit
 #+lispworks4 #+lispworks4 :change-callback-type :item
 #-lispworks4 #-lispworks4 :editing-callback 'record-number-pane-on-edit
                           :visible-min-width '(:character 10)	; see :record-number initarg
                           :visible-min-height #1=(- ywi:*hscroll-bar-height* 2)
                           :font *navigation-font*
			   :internal-border nil
                           :visible-border nil
                           :help-key :name)
           :down :last :new)
  :images '(0 1 nil 2 3 4)
  :callback-type :interface-data
  :selection-callback (lambda (grid data)
                        (pinboard-on-move (pinboard grid) nil nil nil data))
  :font *navigation-font*
  :default-image-set (capi:make-general-image-set
                      :id #.(gp:read-external-image
                             (lw:current-pathname "images/recnavi.bmp")
                             :transparent-color-index 1)
                      :image-count 5
                      :width 10 :height 7)
  :image-width 10 :image-height 7
  :button-width 10 :button-height #1# ;#1=(1+ ywi:*hscroll-bar-height*)
  :visible-min-height 0 :visible-max-height #1#
  :internal-border 1		; DI 2009-Jun-24: top-border for record-number-pane
  :visible-border nil))

(defun %initialize-navigation-bar (self title record-number &optional buttons tooltips)
 ;;; Args: title	 String, T means default title,
  ;;			 NIL or :no-hint means no title; 
  ;;       record-number Explicit min width of record-number-pane in characters.
  ;;       buttons       Additional operation buttons defined as initargs list, e.g.
  ;;			    (:save plist1 :delete plist2)
  ;;			 where plist ::= (:tooltip ... :enabled-function ...)
  ;;       tooltips      Alist of tooltips for standard buttons
  ;;			    ((keyword . tooltip) ...)
  ;; Value: A (freshly consed) list of capi objects to adjoin to the footer layout.
  (with-slots (records-title record-number-pane records-count-title) self
    (setf record-number-pane  (elt (capi:collection-items self) 2)
          records-title       (when (and title (neq title :no-hint))
                                (make-instance 'capi:title-pane
                                 :text (string-append #\Space
                                         (if (stringp title) title #L"Record"))
                                 :font *navigation-font*
                                 :visible-min-height #1=ywi:*hscroll-bar-height*
                                 :visible-border nil))
          records-count-title (make-instance'capi:title-pane :text #L"of "
				:font *navigation-font*
		      		:visible-min-width `(:character ,(1+(or record-number 10)))
                                :visible-min-height #1#
				:visible-border nil))
    (when record-number
      (capi:set-geometric-hint record-number-pane
                               :visible-min-width `(:character ,record-number)))
    (when buttons
      ;; Add component as one but last element of the items and buttons sequences.
      (let ((items (capi:collection-items self)))
        (setf items
              (concatenate 'vector
               (subseq items 0 5)
               (list (make-instance 'capi:toolbar-component
                      :default-image-set (capi:make-general-image-set
                                          :id #.(gp:read-external-image
                                                 (lw:current-pathname "images/recoper.bmp")
                                                 :transparent-color-index 1)
                                          :image-count 2
                                          :width 10 :height 7)
                      :items
                      (loop for rest on buttons by 'cddr
                            for data = (first rest)
                            and value = (second rest)
                            when value
                            collect
                            (make-instance 'capi:toolbar-button
                             :data data
                             :image (navigation-button-image data)
			     :enabled-function (and (consp value)
                                                    (getf value :enabled-function))
                             :tooltip (or (and (consp value) (getf value :tooltip))
                                          (getf (cdr-assoc data *navigation-button-alist*)
                                                :tooltip))))
                      :callback-type :interface-data
		      :selection-callback 'active-grid-operate))
               (subseq items 5))
              (capi:collection-items self) items)		; -> recreate button list
        ;; HACK: Recreated buttons have the image slot set to datum - reassign images!
        (loop ;with image of-type fixnum = -1
              for item across items
              and button in (CAPI::TOOLBAR-BUTTON-GROUP-BUTTONS self)
              when (typep button 'capi:toolbar-button)
              do (setf (capi:toolbar-button-image button)	;(incf image)
                       (navigation-button-image (or (capi:item-data button) item)))
                 (when-let (tooltip (cdr-assoc item tooltips))
                   (setf (capi::toolbar-button-tooltip button) tooltip)) )))
    `(,@(when records-title `(,records-title))
      ,self
      ,records-count-title)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  BASIC-GRID  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Abstract superclass for "true" grids and formularies

(capi:define-interface basic-grid ()
 ((nrows :accessor nrows			; # of rows to display (filtered)
	:type fixnum  :initform 0)		; including the auto-record
  (current-row :accessor current-row :initform 0)
  (records :accessor records			; a la capi:collection-items
	:initform (make-array *records-increment* :fill-pointer 0 :adjustable t))
  (allow-additions :reader allow-additions)	; Integer counters of records have been
  (allow-edits     :reader allow-edits)		; added, changed, or deleted so far.
  (allow-deletions :reader allow-deletions)	; NIL if not allowed
  (addition-maker :reader addition-maker	; Used to create object filled with 
	:initarg :addition-maker		; default values for auto record.
	:initform 'list-addition-maker)		; Signature:(), *grid* is bound during call

  ;;-- Callbacks --
  ;; The following three has signature: (grid record)
  ;; Rules of thumb:
  ;;  - Provide update-callback only if various instances of a ygrid subclass update
  ;;	 differently; if similarly, try to override or extend record-on-update instead.
  ;;  - The update- and delete-callback are only invoked on filled records;
  ;;	 the current-callback can be invoked on an auto-record, so be prepared
  ;;	 to deal with null record-data.
  (current-callback :initarg :current-callback :initform nil)
  (update-callback :reader update-callback :initarg :update-callback :initform nil)
  (delete-callback :initarg :delete-callback :initform nil)

  ;; On changing the grid as a whole the very first time;
  ;; Signature: (top-level-interface grid),  e.g. ywi:interface-on-change
  ;; CAUTION: The :realm property of the grid is never used when calling this 
  ;;	       change-callback from grid functions;
  ;;	       the grid itself is always passed as an argument.
  (change-callback :reader change-callback  :initarg :change-callback  :initform nil))
 (:panes
  (navigation-toolbar navigation-bar :reader navigation-toolbar)))

(defmethod initialize-instance :after ((grid basic-grid) &key (allow-additions t)
                                                              (allow-edits t)
                                                              (allow-deletions t))
  (setf (slot-value grid 'allow-additions) (cond ((integerp allow-additions)
                                                  allow-additions)
                                                 (allow-additions 0))
        (allow-edits grid) allow-edits
        (allow-deletions grid) allow-deletions))

(defun (setf allow-edits) (value grid)
  (setf (slot-value grid 'allow-edits) (cond ((integerp value) value) (value 0))))
(defun (setf allow-deletions) (value grid)
  (setf (slot-value grid 'allow-deletions) (cond ((integerp value) value) (value 0))))

(defmethod retrieval ((grid basic-grid)) :immediate)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  YGRID  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(capi:define-interface ygrid (basic-grid)
 ((ncols :accessor ncols	 		; total number of apparent columns
	:type fixnum  :initform 0)
  (ground-ncols :accessor ground-ncols	 	; total number of ground columns
	:type fixnum  :initform 0)
  (current-col :accessor current-col :initform 0)
  (column-descriptions :initform ()		; either alist of column descriptions
	:initarg :column-descriptions		;  ((name . initargs) ...)
	:reader column-descriptions)		; or function: (interface grid name)->alist
  (columns :initform #() :accessor columns)	; vector of apparent column objects
  (record-constructor			 	; default record constructor
	:initform 'make-record :allocation :class)       ;:initarg :record-constructor
  (row-height :initarg :row-height  :reader row-height  :type fixnum  :initform 0)
  (internal-width :initform nil)		; cache value summing all the columns
  (left-frozen-ncols :reader %left-frozen-ncols	; the number of frozen
	:initarg :left-frozen-ncols :initform 0)	; non-scrollable columns
  (left-frozen-width :initform nil)		; cache value of width of left frozen area
  (gridlines :initarg :gridlines :reader gridlines	; ::= :all | :horizontal 
	:initform :all)					;   | :vertical | nil
  (graphics-args-function			; Signature: (grid record column)
	:reader graphics-args-function
	:initarg :graphics-args-function :initform nil)
  (scroll-x :initform 0 :accessor scroll-x)	; viewport origin coordinates
  (scroll-y :initform 0 :accessor scroll-y)	; can also by modified by hardcopying
  (interaction :accessor choice-interaction	; ::= :no-selection | :single-selection
	:initarg :interaction			;   | :multiple- | :extended-
	:initform :extended-selection)		; a la setfable capi:choice-interaction
  (selected-rows :accessor selected-rows	; The row numbers of selected records
	:initform nil)				; a la capi:choice-selection

  ;; Cell after update; signature: (grid record column &optional recursive)
  ;; Beware of calling (setf cell-value) from inside this callback in case of cycling
  ;; dependencies between columns. This could lead to stack overflow!
  ;; Use (set-cell-value ... :recursive t) or call-data-writer instead.
  (cell-update-callback :initarg :cell-update-callback :initform nil)

  ;; Reuse the activate-callback slot that otherwise has no value for non-top-level grid.
  ;; It must be a function of signature: (grid activatep),
  ;; and could help indicating which grid is current within the interface.
  ;; This "move focus inside grid" hook is called when
  ;; - either the pinboard receives focus,
  ;; - or any editing overlay receives focus,
  ;; - or another grid-component receives a mouse click, so the corresponding method
  ;;   invokes (capi:set-pane-focus pinboard).
  ;; Invoking with activatep NIL is hardly detectable in general;
  ;; so is left at the disposal of the application.
  ;(enter-callback :initarg :enter-callback :initarg :activate-callback :initform nil)

  ;; old value of the current cell; set by cell-on-enter and set-cell-value
  (%old-value% :initform *unbound-value*)
  (workspace-text :reader workspace-text :initarg :workspace-text :initform nil)
  #+lispworks4.3 (%initialized% :initform nil)) 
 (:panes
  (column-header column-header-row :reader column-header
	:font (capi:simple-pane-font capi:interface))
  (row-header row-header-column :reader row-header)
  (hscroll-bar capi:scroll-bar
	:orientation :horizontal
	:visible-border nil			;t  
	:visible-min-width 0
	:accepts-focus-p nil
	:start 0 :end 999			; scroll range
	:line-size *default-column-width*	; scroll increments
	:page-size 50
	:slug-start 0				; initial slug position
	:callback 'horizontal-on-scroll)
  (bottom-right-corner capi:title-pane		; capi::dumm-pane doesn't work
	:visible-border nil
	:visible-min-width 0
        :visible-min-height 0))  ;:visible-max-height 0))
 (:layouts
  (main-layout capi:column-layout '(main-row-layout footer-layout) 
	:y-gap 0)
  (footer-layout capi:row-layout '(navigation-toolbar hscroll-bar bottom-right-corner)
	;:visible-min-height 0
	:x-gap 1 :y-gap 0
	:y-adjust :bottom)
  (main-row-layout capi:row-layout '(row-header main-column-layout) 
	:x-gap 0)
  (main-column-layout capi:column-layout '(column-header pinboard)
	:y-gap 0 :y-adjust :top
	:visible-border nil)
  (pinboard record-board ()
	:reader pinboard
	:font (capi:simple-pane-font capi:interface)))
        ;:background (capi:simple-pane-background capi:interface)
        ;:foreground (capi:simple-pane-foreground capi:interface)
 (:default-initargs
  :layout 'main-layout
  :visible-min-width 0
  ;; NB: Filling cell background manually looks worse that filling in 
  ;;     right-bottom area with background :color_appworkspace 
  ;; The following constrain even a non-top-level grid - so overwrite by
  ;; :visible-min-...  if needed
  :best-width 300  :best-height 200
  ;; The following is pertinent to a grid if it is at top-level itself
  :message-area t
  :initial-focus 'pinboard
  #-lispworks4.3 :create-callback #-lispworks4.3 'interface-on-create
  ;; The following two get only called automatically for top-level grids
  :confirm-destroy-function 'grid-confirm-destroy
  :destroy-callback 'grid-on-destroy))

(defmethod parse-column-descriptions ((grid ygrid) column-descriptions &key reader writer)
 ;; TODO:    ,   -  . ground-col
  (let ((default-ground-col 0)		; index for simple elt-columns
        (apparent-columns ()))
    (declare (fixnum default-ground-col))
    (do* ((rest column-descriptions (rest rest))
          (col 0 (1+ col))
          (description (first rest) (first rest)))
         ((null rest))
      (declare (fixnum col))
      (multiple-value-bind (column ground)
          (make-column (first-or-self description)
                       col
                       (if (consp description) (rest description))
                       :default-ground-col default-ground-col
                       :reader reader  :writer writer)
                         ;(yl:debug-format :grid "column=~a" column)
        (when (and ground (numberp (column-data-reader column)))
          (setf default-ground-col (1+ (max default-ground-col
                                            (column-data-reader column)))))
        (push column apparent-columns)))
    (setf (columns grid) (make-array (length apparent-columns)
                                     :initial-contents (nreverse apparent-columns))
          (ncols grid) (length (columns grid))
          (ground-ncols grid) default-ground-col
          (slot-value grid 'internal-width) nil
          (slot-value grid 'left-frozen-width) nil)
    (columns grid)))

(defmethod initialize-instance :after ((grid ygrid) &key
           (background #-(and lispworks4 win32) :color_window
                       #+(and lispworks4 win32) 'win32:color_window)
           (foreground #-(and lispworks4 win32) :color_windowtext
                       #+(and lispworks4 win32) 'win32:color_windowtext)
           ;header-background header-foreground
           column-header-height
           column-header-font
           row-header-width		; must be specifed >=1 when :visible-border is true
           row-header-align		; horizontal alignment
           (navigation-title t)		; :no-hint or "Title"
           record-number		; :visible-min-width chars for record-number-pane
           (horizontal-scroll t)
           visible-border
           (press-return *default-press-return*)
           select-record		; used along with :allow-edits nil
           cancel-function 		; used by cell-on-cancel
           &allow-other-keys)
  ;;; Initargs:
   ;;  :visible-border
   ;;	 As only :outline seems to have an effect on an embedded capi:interface,
   ;;    we always set the corresponding slot of the grid instance to NIL
   ;;    and further interpret its value as follows:
   ;;	 True  = pass the value to the main layout,
   ;;    False = let row-header and column-header components to render a one-pixel
   ;;            border on the top and left themselves.
   ;;  :cancel-function
   ;;	 Function of signature: (grid)
   ;;	 Useful when the grid is inside a dialog or wizards to emulate the Cancel
   ;;	 button behavior - just specify capi:quit-interface or capi:abort-dialog.
  (with-slots (column-descriptions main-layout
               column-header row-header pinboard
               footer-layout navigation-toolbar hscroll-bar bottom-right-corner)
      grid
    ;; NB: parse-column-descriptions can also be specialized and intended to call later,
    ;;	   from load-grid - this is disharmony!
    ;;	   Hack to prevent subclass methods from being called prematurely.
    (when (consp column-descriptions)
      (parse-column-descriptions grid column-descriptions))
    (setf (capi:simple-pane-background pinboard) background
          (capi:simple-pane-foreground pinboard) foreground)
    ;(when header-background (allf header-background
    ;        (capi:simple-pane-background column-header)
    ;        (capi:simple-pane-background row-header)
    ;        (capi:simple-pane-background navigation-toolbar)
    ;        (capi:simple-pane-background bottom-right-corner)))
    (when row-header-width
      (capi:set-geometric-hint row-header :visible-min-width row-header-width))
    (when row-header-align
      (setf (column-header-align row-header) row-header-align))
    (when column-header-font
      (setf (capi:simple-pane-font column-header) column-header-font))
    (when column-header-height
      (capi:set-geometric-hint column-header :visible-min-height column-header-height))

    ;; We need the following hack to disable :horizontal-scroll t among initargs!
    (setf (slot-value grid 'CAPI-INTERNALS:HORIZONTAL-SCROLL) nil
          (slot-value grid 'CAPI::VISIBLE-BORDER) nil			; not applicable
          (slot-value main-layout 'CAPI::VISIBLE-BORDER) visible-border)
    (allf (null visible-border)
      (slot-value column-header 'border)
      (slot-value row-header 'border)
      (slot-value pinboard 'border))
    (cond ((or navigation-title horizontal-scroll)
           (setf (capi:layout-description footer-layout)		; show if specified
                 `(,.(if navigation-title
                         (%initialize-navigation-bar navigation-toolbar navigation-title
                                                     record-number)
                         (setf navigation-toolbar nil))
                   ,.(if horizontal-scroll
                         `(,hscroll-bar ,bottom-right-corner)
                         (nilf hscroll-bar bottom-right-corner)))))
          (t
           (nilf (capi:layout-description footer-layout)		; hide the entire
                 navigation-toolbar hscroll-bar bottom-right-corner)	; footer-layout 
           (capi:set-geometric-hint footer-layout :visible-max-height 0)))
    ;(initialize-editor-buffer text-input buffer)
    (when press-return
      (setf (capi:capi-object-property grid :press-return) press-return))
    (when select-record
      (setf (capi:capi-object-property grid :select-record) select-record))
    (when cancel-function
      (setf (capi:capi-object-property grid :cancel-function) cancel-function)) ))

;(defmethod capi:pane-initial-focus ((grid ygrid)) ; DI 2010-May-05 replaced by initarg
;  (pinboard grid))

;;; Inlines and Macros

(declaim (inline bottom-exposed-row cell-visible-height col-column
                 %current-column %current-record
                 internal-height internal-width %row-record topmost-exposed-row)
         (ftype (function (t &optional t) fixnum) column-left)
         (ftype (function (t) fixnum)
                topmost-exposed-row bottom-exposed-row
                cell-visible-height internal-width))

(defun col-column (grid col) ; fast column number -> column object
  (svref (columns grid) col))
(defun %row-record (grid row) ; fast row number -> record a la capi:get-collection-item
  (aref (records grid) row))

(defun row-record (grid &optional (row (current-row grid)))
  (declare (fixnum row))
  (if (< -1 row (nrows grid))
      (%row-record grid row)
      nil))

(defun %current-column (grid)
  (col-column grid (current-col grid)))
(defun %current-record (grid)
  (%row-record grid (current-row grid)))

(defun current-column (grid)
  (let* ((columns (columns grid))
         (length (length columns))
         (col (current-col grid)))
    (when (< -1 col length)
      (svref columns col))))

(defun current-record (grid &optional kind)
  (when-let (record (row-record grid (current-row grid)))
    (when (case kind
            (:filled (record-filled-p grid record))
            (:new    (record-new-p grid record))
            (:auto   (record-auto-p grid record))
            (otherwise t))
      record)))

(defmethod name ((grid ygrid))
  (capi:capi-object-name grid))

;;; Read-only column does not allow to save changes to record-data object via gesture.
;;; This macro for internal use only.
(defmacro column-read-only (grid column)
  (declare (ignore grid))
  `(eq (enabled ,column) :read-only))

(declaim (inline column-left-frozen-p column-hidden)
         (ftype (function (t) fixnum) left-frozen-ncols left-frozen-width)
         (ftype (function (t t) boolean) column-left-frozen-p))

(defun left-frozen-ncols (grid)
  (declare #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (the fixnum (min (the fixnum (ncols grid)) (the fixnum (%left-frozen-ncols grid)))))

(defun column-left-frozen-p (grid column)
  (declare #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (< (the fixnum (if (integerp column) column (column-number column)))
     (the fixnum (left-frozen-ncols grid))))		; and (< 0 left-frozen-ncols)?

(defun column-hidden (column)
  (<= (column-width column) 1))

(defun column-width-to-pixels (width grid column)
  (cond ((consp width)
         (case (first width)
           (:character				; (:character integer)
            (+ (* (or (second width) 0)
                  (1+ (gp:get-font-average-width (pinboard grid) ;(gp:get-char-width ..#\m)
                       (getf (column-graphics-args column) :font))))
               4))
           (:string				; (:string string) or (:string :title)
            (let ((string (second width)))
              (+ (if (eq string :title)		; compulte on column header port
                     (gp:port-string-width (column-header grid)
                                           (string (column-title column)))
                     (nth-value 2 (gp:get-string-extent (pinboard grid) string
                                   (getf (column-graphics-args column) :font))))
                 4)))
           (:percentage				; run-time only - initially port-with=>1
            (truncate (* (second width) (gp:port-width (pinboard grid))) 100))
           (symbol-value
            (symbol-value (second width)))
           ((funcall apply)
            (apply (second width) (cddr width)))
           ((max min + - * /)
            (reduce (first width) (rest width)
                    :key (lambda (width) (column-width-to-pixels width grid column))))
           (otherwise
            *default-column-width*)))
        ((numberp width)
         (max width 0))
        (t *default-column-width*)))

(defun %initial-column-pixels (grid column &optional width)
 ;;; Value: Rational number
  (column-width-to-pixels
   (or width
       (when-let (column-descriptions (column-descriptions grid))
         (getf (rest (listify (find (column-name column)
                                    (if (consp column-descriptions)
                                        column-descriptions
                                        (funcall column-descriptions
                                                 (capi:top-level-interface grid)
                                                 grid (ywi:slot-name grid)))
                                    :key #'first-or-self)))
               :width)))
   grid column))

(defun row-height-to-pixels (height grid)
  (cond ((listp height)		; canonicalize row height
         (+ (* (or (second height) 1) (%font-height% (pinboard grid))) 2))
        ((<= height 1)
         (+ (%font-height% (pinboard grid)) 2))
        (t height)))


(defun topmost-exposed-row (grid)
  (the fixnum (values (floor (scroll-y grid) (row-height grid)))))
(defun bottom-exposed-row (grid)
  (the fixnum
       (min (the fixnum
                 (values (ceiling (+ (scroll-y grid)
                                     (capi:simple-pane-visible-height (pinboard grid)))
                                  (row-height grid))))
            (the fixnum (1- (nrows grid))))))

(defun cell-current-p (grid row-or-record column)
  (and (eql (if (integerp row-or-record) row-or-record (record-row-number row-or-record))
            (current-row grid))
       (eql (column-number column) (current-col grid))))

(defun cell-visible-height (grid)
  (case (gridlines grid)
    ((:all :horizontal)
     (the fixnum (1- (row-height grid))))
    (otherwise
     (row-height grid))))
     
(defun internal-height (grid)
  (* (nrows grid) (row-height grid)))		; may not be a fixnum!

(defun internal-width (grid)
  (with-slots (internal-width) grid
    (the fixnum (orf internal-width (reduce #'+ (columns grid) :key 'column-width)))))

(defun left-frozen-width (grid)
 ;;; Total width of left frozen columns
  #-debug (declare (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (with-slots (left-frozen-width) grid
    (the fixnum (orf left-frozen-width
                     (do ((width 0)
                          (columns (columns grid))
                          (left-frozen-ncols (left-frozen-ncols grid))
                          (col 0 (1+ col)))
                         ((>= col left-frozen-ncols)
                          width)
                       (declare (fixnum width left-frozen-ncols col)
                                (type simple-vector columns))
                       (incf width (the fixnum (column-width (svref columns col)))))))))

(defmacro with-cell-mask ((mask-left mask-width) grid column left width &body body)
 ;;; Calculate the horizontal parameters of the mask to apply later during drawing
 ;;; within the body. Take the frozen columns into account.
  ;; Args: mask-left, mask-width
  ;;			Variables that are bound around the body
  ;;       column,width Arbitrary lisp forms
  ;;       left         Internal coordinate (should be a variable)
  ;;	   grid 	Should be a variable
  ;; Values bound:
  ;;  mask-left  An internal coordinate
  ;;  mask-width Can be negative!
  ;; NB: These can differ from left an width only for non-frozen columns!
  (with-gensyms (new)
   `(let ((left-frozen-ncols (left-frozen-ncols ,grid))
          (,mask-left ,left)
          (,mask-width ,width))
      (declare (fixnum left-frozen-ncols ,mask-left ,mask-width))
      (when (and (< 0 left-frozen-ncols)		; the column is not frozen
                 (<= left-frozen-ncols (column-number ,column)))
        (let ((,new (+ (scroll-x ,grid) (left-frozen-width ,grid))))
          (declare (fixnum ,new))
          (when (< ,left ,new)				; decrease mask width
            (setq ,mask-width (the fixnum (- ,mask-width (the fixnum (- ,new ,left))))
                  ,mask-left ,new)
            (when (< ,mask-width 0)			; ensure positive mask width
              (setq ,mask-width 0)))))
      ,@body)))

(defun column-left (grid &optional (column (%current-column grid)))
 ;;; Value: Semi-internal left of the columm given
  (the fixnum (+ (the fixnum (loop for c across (columns grid)
                                   until (eq c column)
                                   sum (column-width c)))
                 (the fixnum (if (column-left-frozen-p grid column) (scroll-x grid) 0)))))

(defmacro record-cache-value ((grid record column) form)
 ;;; Obtain a value to be drawn from cache or by calling data-reader
  ;; Args: form  Must refer to %datum% and %condition% lexical.
  ;; Called by: Methods on draw-cell or the like that should quickly get printed
  ;;		representation of the column value without heavy recomputation.
  (with-gensyms (cache cache-vector col)
    `(let (,cache
           (,cache-vector (record-%cache% ,record))
           (,col (column-number ,column)))
       (declare (fixnum ,col))
       (if (eq (setq ,cache (svref (or ,cache-vector
                                       (setq ,cache-vector	; install new cache
                                             (setf (record-%cache% ,record)
                                                   (make-array (ncols ,grid)
                                                     :initial-element *unbound-value*))))
                                   ,col))
               *unbound-value*)
           (multiple-value-bind (%datum% %condition%)	;cell value
               (ignore-errors (call-data-reader ,grid (record-data ,record) ,column)) 
             (setf (svref ,cache-vector ,col) ,form))
           ,cache))))

#|(defun record-cache-svref (record column &optional default)
 ;;; Value: An element of the cache vector or the default given if cache has not been
  ;;	    filled yet.
  ;; NB: Unused in cell-graphics-args function as it is called before record-cache-value
  (if (and (setq record (record-%cache% record))
           (neq (setq record (svref record (column-number column))) *unbound-value*))
      record
      default))|#

(defun selected-columns (grid)
 ;;; loop results in code of double size.
  ;(loop for column across (columns grid) when (column-selected column) collect column))
  (declare (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (do* ((columns (columns grid))
        column
        (acc ())
        (i (1- (the fixnum (length columns))) (1- i)))
       ((< i 0) acc)
    (declare (type simple-vector columns) (fixnum i))
    (when (column-selected (setq column (svref columns i)))
      (setq acc (cons column acc)))))

;;; "Being edited" status of the current record
(defmacro %get-edit-mode (grid)
  `(capi:capi-object-property ,grid :edit))

(defun %set-edit-mode (grid row-or-record value &key (redraw-now t))
 ;;; Args: row-or-record Row number (integer) only when invoked from methods on formulary.
  ;;       enter NIL only when invoked by formulary, e.g. %fill-auto-record
  (when (xor value (capi:capi-object-property grid :edit))
    (if value
        ;(progn
        ; (when (record-auto-p grid row-or-record)       ; user fills top left cell 
        ;    (if enter				         ; ywi:commit->write-displayer-data
        ;       (record-on-enter grid row-or-record)     ; fill and invoke current-callback
        ;       (%fill-auto-record grid row-or-record))) ; fill not invoking it
	;; Set "being edited" mode only if OK
        (record-on-change (capi:top-level-interface grid) grid row-or-record)
        ;; Drop "being edited" mode
        (setf (capi:capi-object-property grid :edit) nil))
    ;(when (and record (= (record-row-number record) (current-row grid)))
    (when redraw-now
      (when-let (row-header (row-header grid))
        (draw-cell grid (or row-or-record (%current-record grid)) row-header nil nil)))))

;;; Analogy to capi:choice
;(defun collection-items (grid) (records grid))
;(defun count-collection-items (grid) (length (collection-items grid)))

(defmacro do-records ((record grid &key (filter :all)) &body body)
 ;;; Args: filter ::= :all Iterate over both shown and filtered out records-off
  ;;		    | :on  Only over those that are on
  ;;		    | :off Only over those that are off
  (lw:rebinding (grid filter)
    (with-gensyms (records records-off nrows end row)
      `(do* (,record
             (,records (records ,grid))
             (,records-off (if (eq ,filter :on) () (slot-value ,grid 'records-off)))
             (,nrows (nrows ,grid))
             (,end (cond ((eq ,filter :off) 0)
                         ((record-auto-p ,grid (1- ,nrows)) (1- ,nrows))
                         (,nrows)))
             (,row 0 (1+ ,row)))
            ((and (>= ,row ,end) (null ,records-off)))
         (declare (fixnum ,nrows ,end ,row))
         (setq ,record (if (< ,row ,end) (aref ,records ,row) (pop ,records-off)))
         ,@body))))

(defun reduce-grid (function grid &key initial-value key (filter :on))
 ;;; A la reduce, funcall the function only over non-nil data.
  ;; Args: key Applied to record-data
  (do-records (record grid :filter filter)
    (when-let* ((data (record-data record))
                (value (if key (funcall key data) data)))
      (setq initial-value (if initial-value
                              (funcall function initial-value value)
                              value))))
  initial-value)

(defvar *background-cleared* nil
  "Used by DRAW-CELL. Top-level redraw procedure rebind it to T after clear-rectangle.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  GENERICS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric draw-cell (grid record column port position &optional graphics-args)
 ;;; NB1: Graphic translation is applied by methods itself, not by the caller.
  ;; NB2: Normally, draws *unbound-value* as empty cell.
 (:documentation
  "Draw cell contents on PORT.
Arguments:
 RECORD : record or header row port,
 COLUMN : apparent or header column,
 POSITION : either rectangle of a cell interior in internal coordinates
	or x or y in internal or visual coordinates."))

(defgeneric compute-natural-size (grid component)
 (:documentation
   "Calculate best-fit width of column or height of row/record argument.
- Should subsume padding.
- Only good (non-signaling) data is counted.")
 (:method ((grid ygrid) (column column))
  ;; Restore the column to default extracted from the column-description
  (%initial-column-pixels grid column)))

(defvar %recursive% nil)		; if true, we can only go outwards, not inwards

(defgeneric grid (object)
  (:documentation "Returns the single or current (if several) basic-grid instance
which menu and toolbar items operate on inside the interface or other object.")
  (:method (object) nil)
  (:method ((object CAPI::BASIC-ELEMENT))
   (when (CAPI::PANE-CREATED-P object)
     (grid (capi:element-interface object))))
  (:method ((interface capi:interface))
   (cond ((capi:top-level-interface-p interface)
          (unless %recursive%
            (let ((%recursive% t))
              (grid (capi:pane-descendant-child-with-focus interface)))))
         ((slot-value interface 'CAPI::CREATED)
          (grid (capi:element-interface interface)))))
  (:method ((layout capi:layout))		; the layout should contain a grid
   (if %recursive%
       (grid (capi:element-interface layout))	; move outwards from focus
       (ywi:find-pane layout 'basic-grid :leaf-only nil)))
       ;(let ((%recursive% t))
       ;   (or (grid (capi:pane-descendant-child-with-focus layout))  ; !
       ;        (ywi:find-pane layout 'basic-grid :leaf-only nil))))))
  (:method ((grid basic-grid)) grid) )

(defgeneric operate-on-grid (interface grid operation)
 (:documentation "Called from edit/record toolbar/menu as selection-callback.")
 (:method (interface grid operation)
  (beep-not-implemented interface operation))

 (:method :around (interface (grid null) operation)
  (when (setq grid (grid interface))			; to prevent looping
    (operate-on-grid interface grid operation)))

 (:method (interface (grid basic-grid) (operation function))
  (declare (ignore interface))
  (funcall operation grid))

 (:method (interface (grid basic-grid) (operation cons))
  (declare (ignorable interface))
  (let ((second (second operation)))
    (if (and (eq (first operation) 'function) (symbolp second) (fboundp second))
        (funcall second grid)
        (call-next-method)))) )


(defgeneric operate-on-grid-p (interface grid operation)
 (:documentation "Predicate returning true if an operation is possible.
Called from edit/record toolbar/menu as enabled-function.")
 (:method (interface grid operation)
  t)
 (:method :around (interface (grid null) operation)
  (if (setq grid (grid interface))			; to prevent looping
      (call-next-method)
      nil)))

;;; Forward declarations

(declaim (ftype (function (t t t  fixnum t fixnum fixnum &optional t t  t) (values))
                draw-wrapped-text)
         (ftype (function (t t t t  fixnum fixnum &optional t t) t)
                wrap-column-text))
