;;;-*- Mode: Lisp; -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Check-column functionality
;;;
;;; Values 0 and :false are displayed as unchecked
;;; When nullable is true, provide triple-valued logic:
;;;   if nullable eq :number, cicle over NIL -> 1 -> 0 -> NIL
;;;   otherwise, cicle over NIL -> :true -> :false -> NIL

(in-package :ystok.grid)

(defparameter *check-button-size* 10)  ; height=size, width=size+1

(defconstant +box-size+ 12)
(defconstant +check-box-size+ 11)
(defvar *check-radio-boxes-external-image*
  #.(gp:read-external-image (merge-pathnames "images/check-radio-boxes.bmp"
                                        (or *compile-file-truename* *load-truename*))
                            :transparent-color-index 1))
(defvar *check-radio-boxes-external-image-inversed*	; for foreground coloring
  #.(gp:read-external-image (merge-pathnames "images/check-radio-boxes.bmp"
                                        (or *compile-file-truename* *load-truename*))
                            :transparent-color-index 0))

;; Images to be drawn on ordinary non-selected cell in standard (or null) foreground color
(defvar check-radio-boxes-image nil)
(defun check-radio-boxes-image (port)
  (if (or (null check-radio-boxes-image)
          (gp:image-freed-p check-radio-boxes-image))
      (setq check-radio-boxes-image
            (gp:convert-external-image port *check-radio-boxes-external-image*))
      check-radio-boxes-image))

(defclass check-column (column)
 ((header-align :initarg :align)
  (nullable :initform nil)		; if true, provide triple-valued logic
  (sort-asc :initform 'logical<)
  (sort-desc :initform 'logical>)))
 ;(:default-initargs :header-align :center)

(defmethod draw-cell (grid record (column check-column) port rect
                      &optional (graphics-args (cell-graphics-args grid record column)))
 ;;; Args: graphics-args
  ;;		When foreground color is given, it replaces transparent color
  ;;		in the external image, and we cache the resulting image.
  ;(with-slots (scroll-x scroll-y) grid
  (declare (fixnum *check-button-size*)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (rect-bind (left top width height) rect
    (declare (fixnum left width))
    (let ((left-frozen-ncols (left-frozen-ncols grid))
          (mask-left left)
          (mask-width width))
      (declare (fixnum left-frozen-ncols mask-left mask-width))
      (when (and (plusp left-frozen-ncols) (<= left-frozen-ncols (column-number column)))
        (let ((new (+ (scroll-x grid) (left-frozen-width grid))))
          (declare (fixnum new))			; check if we must make mask narrow
          (when (> new left)
            (setq mask-width (max (- mask-width (- new left)) 0)
                  mask-left new))))
      (let ((mask (translate-mask port rect :left mask-left :top top :width mask-width))
            (value (record-cache-value (grid record column) 
                     (if %condition% 
                         nil				; mixed-value (undetermined) state
                         (case %datum%
                           ((0 :false) :false)
                           ((nil) (if (column-nullable column) nil :false))
                           (otherwise :true))) )))
        (gp:with-graphics-mask (port mask)
          (let ((x (the fixnum (+ left (the fixnum
                  (max (the fixnum (ash (the fixnum (- width +check-box-size+)) -1)) 1)))))
                (y (+ top (max (1- (ash (- height +check-box-size+) -1)) 1)))
                (background (getf graphics-args :background))
                (foreground (getf graphics-args :foreground)))
            (declare (fixnum x))		 ; y - could be bignum?
            (cond (background
                   (gp:draw-rectangles port rect :filled t :foreground background))
                  ((not *background-cleared*)
                   (apply #'gp:clear-rectangle port mask)))
            (gp:draw-image port
                           (if foreground
                               (gp:with-graphics-state (port :background foreground)
                                 (gp:convert-external-image port
                                         *check-radio-boxes-external-image-inversed*
                                         :cache t))
                               (check-radio-boxes-image port))
                           x y
                           :from-x (case value
                                     (:false 0)
                                     (:true +box-size+)
                                     (otherwise (+ +box-size+ +box-size+)))
                           :from-width +check-box-size+
                           :to-width +check-box-size+  :to-height +check-box-size+
                           :operation (if (or (record-selected record)
                                              (column-selected column))
                                          boole-orc1 boole-1)
) ) ) ) ) ) )

(defun toggle-check (value &optional nullable)
  (if nullable
      (case value			; nil => :true => :false => nil
        ((nil)
         (if (eq nullable :number) 1 :true))
        ((:false 0) nil)
        (otherwise			; positive number is treated as 1
         (if (iplusp value) 0 :false)))	; 1 = 2 =...= :true => 0 = :false
      (case value			; nil = 0 = :false => t = :true
        ((nil) t)
        (:false :true)
        (:true :false)
        (0 1)
        (1 0))))
        ;(otherwise nil))))

(defmethod cell-on-character ((grid ygrid) (column check-column) (character (eql #\Space)))
 ;;; Invoked on typing a character when the pinboard has focus.
  ;; The column is guaranteed to be enabled, but may be read-only.
  (if (column-read-only grid column)
      (beep)
      (if-bind (record (current-record grid))
        (multiple-value-bind (datum condition)
            (ignore-errors (call-data-reader grid (record-data record) column))
          (cond (condition (beep))
                ((not (save-cell grid record column
                                :value (toggle-check datum (column-nullable column))))
                 (ensure-cell-visible grid
                                      (record-row-number record) (column-number column)))))
        (beep))))

(defmethod cell-on-click ((grid ygrid) row (column check-column) x y)
 ;;; Toggle current data value
  (when (allow-edits grid)
    (let ((record (%row-record grid row)))
      (multiple-value-bind (datum condition)
          (ignore-errors (call-data-reader grid (record-data record) column))
        (cond (condition (beep))
              ((not (save-cell grid record column
                               :value (toggle-check datum (column-nullable column))))
               (ensure-cell-visible grid row (column-number column))) )))))

(defmethod cell-on-exit (grid record (column check-column))
  t)

(defmethod save-cell ((grid ygrid) (record record) (column check-column)
                       &key (value *unbound-value*) (redraw-now :header))
 ;;; Saving check cell occurs immediatly on click or press.
  ;; Args: value Explicit value passed by cell-on-character/click, set-cell-clip,
  ;;		 or undo-cell
  (if (eq value *unbound-value*)	; no need to save if no value has been specified
      (values t t)
      (restart-case (handler-bind ((error #'prompt-for-restart))
                      (set-cell-value grid record column value :redraw-now redraw-now)
                      (values t nil))
        (abort () 
         :report (list #L"(Cancel) Ignore toggling the check cell at row ~d column ~s."
                       (1+ (record-row-number record))
                       (column-title column))
         (values nil :abort)))))

;;; Edit/select

(defmethod cell-clip-formats ((column check-column))
  '(:value :string))

(defmethod cell-clip ((grid ygrid) (record record) (column check-column))
  ;; NB: Returning "True"/"False"
  ;;	 - accepted only in Excel, not by Access
  ;;	 - can by locale-dependent
  (multiple-value-bind (datum condition)
      (ignore-errors (call-data-reader grid (record-data record) column)) 
    (if condition
        nil
        (values (case datum
                  ((0 :false) "0") 					 	; "False"
                  ((nil) (if (column-nullable column) +null-string+ "0"))	; "False"
                  (otherwise "1"))						; "True"
                (list :value datum)))))

(defmethod set-cell-clip ((grid ygrid) (record record) (column check-column)
                          format (text string)
                          &key (redraw-now :header))
  (save-cell grid record column  :redraw-now redraw-now
             :value (cond ((string= text "0") 0)
                          ((string-equal text "FALSE") :false)
                          ((zerop (length text))
                           (if (column-nullable column) nil :false))
                          (t :true))))
	                  	;((member text '("1" "-1") :test 'string=) 1)
        	          	;((string-equal text "TRUE") :true)

(defmethod set-cell-clip ((grid ygrid) (record record) (column check-column)
                          (format (eql :value)) value
                          &key (redraw-now :header))
  (if (stringp value)
      (call-next-method)
      (save-cell grid record column  :redraw-now redraw-now
                 :value (if (member value '(1 0 :false t nil)) value :true))))

#|(defmethod clipboard-value-bad (grid (column check-column) &optional record)
 ;;; Any value fits but convert "arbitrary true" to logical equivalent
  (let ((datum (capi:clipboard grid :value)))
    (if (member datum '(1 0 :false t nil)) datum :true)))|#

(defmethod operate-on-cell-p ((grid ygrid) (record record) (column check-column)
                              (operation (eql :copy)))
  (enabled column))				; there is no :around method for :copy

;(defmethod operate-on-cell ((grid ygrid) (record record) (column check-column)
;                            (operation (eql :copy)))
; (capi:set-clipboard grid (cell-value grid record column) (cell-text grid record column)))

(defmethod operate-on-cell-p ((grid ygrid) (record record) (column check-column)
                              (operation (eql :paste)))
  (not (and (capi:clipboard-empty grid :value) (capi:clipboard-empty grid :string))))

(defmethod operate-on-cell-p ((grid ygrid) (record record) (column check-column)
                              (operation (eql :undo)))
  (and (enabled column) (neq (slot-value grid '%old-value%) *unbound-value*)))

(defmethod undo-cell ((grid ygrid) (record record) (column check-column))
 ;;; Full revert: since we invoke save-cell on every change, exploit the same to rollback.
  (with-slots (%old-value%) grid
    (unless (eq %old-value% *unbound-value*)
      (if (eql %old-value% (call-data-reader grid (record-data record) column))
          (beep)
          (save-cell grid record column :value %old-value%))
      (%after-revert-cell grid record column t))))
