;;;-*- Mode: Lisp; -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Example interface with two embeded grids
;;;
;;; LWW 6.1.1 BUG:
;;;   When the initial value of row-height of any grid is too small,
;;;   LW signals on clicking within any text cell.
;;;   This is due to a CAPI editor-pane bug.
;;;   After increasing the row height by means of (setf grid:row-height)
;;;   or dragging the row header divider, everything works fine.

(defun lp-col3-reader (data-sequence)
  (> (elt data-sequence 1) 5))

(defparameter *bold-font* (gp:gf nil "helvetica" :bold :roman 8))

;;; There are 7 columns total but only 6 are ground.
;;; "Col 3 SQR" is a kind of ghost, or computed, column as its data-writer is null.
(defparameter *column-descriptions*
 `((col1  :title "ID 1"
          :align :center
          :width 30
          :cell-action-callback demo-cell-action-callback)
   (col2  :title "Float 2"
          :align :right
          :width (:character 10)
          :print-function ywi:float-princ-to-string-d2		; a la "~,2f"
          :data-read-converter yl:e-format-to-string
          :data-write-converter ywi:real-write-converter
          :sort-asc grid:nullable<  :sort-desc grid:nullable>)
   ("Col 3 SQR"
          :align :right  :header-align :center
          :help-key "This column value is [Float 2] * [Float 2]"
          ;:data-reader col3-reader
          :data-reader ,(load-time-value (compile nil
                         '(lambda (data) (declare (ignore data))
                            (when-let (col2-value (grid:cv col2))
                              (* col2-value col2-value)))))
          :data-writer nil
          :print-function yl:$-format-to-string 
          :data-read-converter yl:$-format-to-string
          :sort-asc grid:nullable<  :sort-desc grid:nullable>
          :graphics-args (:background :red))
   (col4  :title "Col 4 one-click sorting"
          :align :center
          :width (:character 20)
          :nullable nil
          :completion-function demo-cell-completion-function
          :header-click-callback grid:reverse-sorting
          :sort-asc lw:locale-string-lessp :sort-desc lw:locale-string-greaterp)
   (col5  :type :check  :title "Chk 5"
          :width 20
          :cell-action-callback demo-cell-action-callback)
   (col6  :title "Currency 6"
          :align :right
          :print-function ,(lambda (arg) (yl:$-format-to-string arg :symbol t))
          :data-read-converter yl:f-format-to-string
          :data-write-converter ywi:float-write-converter
          :test-function col6-float-test)
   (col7  :type :check
          :title "Chk 7"
          :width 35
          :data-writer col7-writer)
   (col8  :type :display  :title "Display 8"
          :cell-action-callback demo-cell-action-callback)
) )

#|(compile 'col3-reader '(lambda (data) (declare (ignore data))
                           (when-let (col2-value (grid:cv col2))
                             (* col2-value col2-value))))|#


(defun col6-float-test (x y)
  (<= (abs (- x y)) 0.1))

;;; Ghost column reader

(defun col7-writer (value sequence)
  (if value
      (setf (elt sequence 5) value)
      (error "Only 'yes' allowed for column 'Chk 7'")))

(defun demo-addition-maker ()
  (let ((id (random 100)))
    (values (list* id (+ id 11.) (make-list (- (grid:ground-ncols grid::*grid*) 2)))
            '(0 1))))			; Declare col1 and col2 as modified

(defun demo-graphics-args (grid record column)
  (let* ((col (grid:column-number column))
         (row (grid:record-row-number record))
         (data (grid:record-data record))
         (value (ignore-errors (grid:call-data-reader grid data column))))
    (case col
      (1 (if (and (numberp value) (> value 10)) (list :foreground :red) nil))
      (2 (if (evenp row) (list :background :red) nil))
      (4 (if value (list :foreground :green) nil))
      (5 (if (and data (elt data 3)) (list :font *bold-font*) nil)) )))

;;; Callbacks on cell

(defun demo-cell-update-callback (grid record column &optional recursive-p)
  (ywi:message-area-format grid "After updating ~S cell: row index ~S column ~S"
                           (capi:capi-object-name grid)
                           (grid:record-row-number record)
                           (grid:column-title column)))

(defun demo-cell-action-callback (grid record column)
  (capi:display-message "Action on ~S cell: record ~S column ~S"
                        (capi:capi-object-name grid) record (grid:column-title column)))

(defun demo-cell-completion-function (grid record column string)
  (string-append (if (> (length string) 4)
                     (subseq string 0 4)
                     string)
                 (format nil "~D" (random 10000))))

;;; Calbacks on record

(defun lisp-value-to-sql (value &optional column)
  (typecase column
    (grid:check-column
     (if value "True" "False"))
    (t (if value (prin1-to-string value) "Null"))))

(defun demo-current-callback (grid record)
  (ywi:message-area-format grid "Grid current: record ~S"  record))

(defun demo-delete-callback (grid record)
  (let ((id (elt (grid:record-data record) 0)))
    (cond ((evenp id)
           (error "Can't delete a row with even ID value."))
          (t
           (format t "~&DELETE * FROM ~a WHERE ID=~d;" "DEMO_TABLE" id)))))

(defun demo-update-callback (grid record)
  (let ((id (elt (grid:record-data record) 0))
        (changed-cv-alist (ywi:changed record)))
    (cond ((null changed-cv-alist))
          ((getf (grid:record-plist record) :new)
           (let ((ground-columns
                  (remove "Col 3 SQR" (grid:columns grid) ; remove pseudo-column
                          :key #'grid:column-name
                          :test #'equalp)))
             (format t "~&INSERT INTO ~a (~{~a~^, ~}) VALUES (~{~a~^, ~});"
                     "DEMO_TABLE"
                     (map 'list #'grid:column-name ground-columns)
                     (map 'list #'lisp-value-to-sql
                          (grid:record-data record)
                          ground-columns)))
           (remf (grid:record-plist record) :new)
           (remf (grid:record-plist record) :changed))
          ((oddp id)
           (error "Can't update the row with odd ID value."))
          (t
           (let ((av-pairs
                  (mapcar (lambda (pair)
                            (destructuring-bind (column . old-value) pair
                              (list (grid:column-name column)
                                    (lisp-value-to-sql (grid:call-data-reader
                                                        grid (grid:record-data record)
                                                        column)
                                                       column))))
                          changed-cv-alist)))
             (when av-pairs
               (format t "~&UPDATE ~a SET ~:{~a=~a~:^, ~} WHERE ID=~d;"
                       "DEMO_TABLE" av-pairs id)
               (remf (grid:record-plist record) :changed)))) )))

(capi:define-interface demo-grid-face ()
 ()
 (:panes
  (a-la-mclp grid:ygrid
   :title "Tick instead of select:"  :title-position :top 
   :column-descriptions
     '(("C1" :type :check :width 20			;(:percentage 20)
             :cell-action-callback demo-cell-action-callback)
       ("LT2" :align :left :width (:string :title)	;(:percentage 50)
              :help-key "a la multi-column list panel"
              :cell-action-callback demo-cell-action-callback)
       ("RT3" :align :right :width (:character 10)
              :cell-action-callback demo-cell-action-callback))
   :gridlines :vertical
   :column-header-font (gp:make-font-description :family "arial"
                                                 :weight :bold  :slant :roman  :size 8)
   :row-height 18
   :row-header-width 0
   :navigation-title nil
   :horizontal-scroll nil
   :visible-border t
   ;:interaction :single-selection
   :allow-edits t
   :allow-additions nil
   :press-return :action)
   ;:select-record t		; highlight entire row
   ;:foreground :blue
  (grid grid:sorted-grid
   :reader grid:grid		; all toolbar/menu commands are pertinent to grid
   :title "Extended selection with one non-scrollable column, localized number and currency format:"
   :title-position :top
   ;:title-args '(:visible-min-height (:character 3))
   :column-descriptions *column-descriptions*
   :addition-maker 'demo-addition-maker
   :graphics-args-function 'demo-graphics-args
   :font (gp:gf nil "arial" :normal :roman 8)
   :left-frozen-ncols 1
   ;:column-header-height 30
   :cell-update-callback 'demo-cell-update-callback
   :current-callback 'demo-current-callback
   :delete-callback 'demo-delete-callback
   :update-callback 'demo-update-callback
   :workspace-text "Put description of demo-version limitations here"))
 (:layouts
  (main-column-layout  capi:column-layout
   '(top-docking main-row-layout))
  (top-docking  capi:docking-layout ()
   :items '()
   :divider-p t)            ; Separator between the toolbar and the menubar
  (main-row-layout  capi:row-layout 
   '(a-la-mclp :divider grid)
   :x-ratios '(1 nil 4)) )
 (:default-initargs
  :title (format nil "YstokGrid Demo on LispWorks ~a" (lisp-implementation-version))
  :layout 'main-column-layout
  :message-area t 
  :best-width 700 :best-height 300
  :confirm-destroy-function 'grid:grid-confirm-destroy
  :help-callback 'ywi:help-callback))

(defmethod initialize-instance :after ((self demo-grid-face) &key &allow-other-keys)
  (with-slots (top-docking) self
    (setf (capi:docking-layout-items top-docking)
          (list (ywi:make-edit-toolbar) (grid:make-records-toolbar))))
  (ywi:insert-menu 'ywi:edit-menu self)
  (ywi:insert-menu 'grid:records-menu self))

(defmethod capi:interface-display :after ((face demo-grid-face))
  (fill-grid (grid:grid face) 20))
        
(defmethod grid:operate-on-grid :after ((face demo-grid-face) grid operation)
  (with-slots (top-docking) face
    (ywi:update-toolbars top-docking 'grid:records-toolbar)))

(defun fill-grid (grid nrows &optional text wipe)
  (when wipe
    (grid:wipe-grid grid))
  (dotimes (row nrows)
    (grid:add-record grid
       (loop for col upfrom 0 below (grid:ncols grid)
             unless (member col '(2))
             collect (case col
                       (0 (1+ row))
                       (1 (+ row 1.11))
                       (4 (oddp row))
                       (5 (+ (* row 100.0) 0.5))
                       (otherwise
                        (format nil "Row ~A by Column ~A" (1+ row) (1+ col))))))))

(defun fill-mclp (grid nrows &optional text wipe)
  (when wipe
    (grid:wipe-grid grid))
  (dotimes (row nrows)
    (grid:add-record grid
                     (list (if (oddp row) :true :false)
                           (format nil "~A left" row)
                           (format nil "right ~A" row)))))


(defun test-demo-grid ()
  (let ((demo (make-instance 'demo-grid-face)))
    (fill-mclp (slot-value demo 'a-la-mclp) 10)
    (capi:display demo)
    demo))


#||;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION transcript ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(setq test (test-demo-grid))
(setq test (make-instance 'grid-test))
(setq g (slot-value test 'grid)
      a-la-mclp (slot-value test 'a-la-mclp))
(fill-grid g 100)
(capi:display test)
(capi:quit-interface g)
(capi:destroy test)
(setf (capi:capi-object-property a-la-mclp :select-record) t) ; select entire row

(setf (grid:row-height a-la-mclp) 15)
||#

