;;;-*- Mode: Lisp; -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Records toolbar/menu (mostly for sorted-grid), edit toolbar/menu

(in-package :ystok.grid)

(defun record-image-set ()
  (capi:make-general-image-set :id
    #.(gp:read-external-image (lw:current-pathname #-Russian "images/record.bmp"
                                                   #+Russian "images/record-ru.bmp")
                              :transparent-color-index 13)
    :image-count 19  :width 16  :height 16))
(define-symbol-macro *record-image-set* (record-image-set))

;;; Generic edit. More bulk operations are implemented in record-clipboard.lisp

(defmethod operate-on-grid-p (interface (grid ygrid) (operation symbol))
  (if (memq operation '(:cut :copy :paste :undo :select-all :zoom))
      (and (not (or (selected-rows grid) (selected-columns grid)))
           (operate-on-cell-p grid (current-record grid) (current-column grid) operation))
      (call-next-method)))

(defmethod operate-on-grid (interface (grid ygrid) (operation symbol))
  (case operation
    ((:cut :copy :paste :undo :select-all)
     (if (or (selected-rows grid) (selected-columns grid))
         (beep-not-implemented interface (format nil "Bulk ~a" operation))
         (operate-on-cell grid (current-record grid) (current-column grid) operation)))
    ;(new-record	    (new-record grid))
    ;(save-record	    (save-record grid))
    ;(delete-record	    (delete-record-or-rows grid))
    ;(:revert-record	    (revert-record grid))
    (:insert		    (insert-record grid))
    ((:sort-asc :sort-desc) (sort-records grid operation nil))
    (:filter-by		    (filter-by-selection grid (%current-column grid)))
    (:toggle-filter	    (toggle-filter grid))
    (:destroy-filter	    (destroy-filter grid))
    (:select-record         (select-section grid (current-row grid) t))
    (:select-all-records    (select-section grid :all))
    ;(searcher	    (capi:find-interface 'searcher :grid grid))
    (otherwise		    (call-next-method)) ))		; DI 2009-Jun-30

(defmethod operate-on-grid (interface grid (operation (eql :refresh)))
  (operate-on-interface interface :refresh grid))

(defgeneric record-on-insert (grid record)
 (:documentation "Invoked after the application has insertd record content successfully.
Value: True, if succeeded.")
 (:method ((grid ygrid) (record record))
  (let ((allow-additions (incf (slot-value grid 'allow-additions))))
    (declare (integer allow-additions))
    (when (and (= allow-additions 1)
               (not (iplusp (allow-deletions grid)))
               (not (iplusp (allow-edits grid))))
      (when-let (change-callback (change-callback grid))
        (funcall change-callback (capi:top-level-interface grid) grid))))
  t))

(defgeneric insert-record (grid)
 ;;; Insert and empty record before the current
  ;; Value: New record instance or NIL if failed.
 (:method :around ((grid ygrid))
  (let ((current-record (current-record grid :filled)))
    (cond ((null current-record)
           (beep))
          ((not (allow-additions grid))
           (beep (list #L"Record additions are not allowed within grid~@[ '~a'~]."
                       (object-title grid))
                 grid))
          ((record-on-exit grid current-record :redraw-now nil)	; -> save-record->save-cell
           (when-let (new-record (call-next-method))
             (invalidate-grid grid)				;:redraw-now :all
             (when-let (current-callback (slot-value grid 'current-callback))
               (funcall current-callback grid new-record)))))))

 (:method ((grid ygrid))
  (let* ((current-row (current-row grid))
         (new-record (funcall (slot-value grid 'record-constructor)
                              :row-number current-row)))
    (when (and (%fill-new-record grid new-record)
               (restart-case
                   (handler-bind ((error #'prompt-for-restart))
                     (record-on-insert grid new-record))
                 (abort ()
                        :report #L"(Cancel) Abort record insertion."
                        nil)))
      (let ((records (records grid)))
        (vector-push-extend nil records *records-increment*)	; shift records in the
        (loop for row downfrom (nrows grid) above current-row	; vector below the current
              for record = (aref records (1- row))
              do (setf (record-row-number record) row
                       (aref records row) record))
        (incf (nrows grid))
        (setf (aref records current-row) new-record))))) )	; set new as current



(defun get-interface-active-grid (object)
 ;;; Helper called by active-grid-* and others.
  ;; Args: object Either focus pane supplied by a selection-callback
  ;;              or top-level interface supplied by enabled-function.
  ;; Values: 1) Top-level interface of the object,
  ;;	     2) Current grid; the focus pane may reside outside it or even
  ;;		belong to a different grid!
  (let* ((interface (if object
                        (capi:top-level-interface object)
                        (current-top-level-interface)))
         (grid (if (typep object 'basic-grid)
                   object
                   (grid (or interface object)))))		; start peeking from top
    (yl:debug-format :grid "<< get-interface-active-grid ~s ~s" grid interface)
    (values interface grid)))

(defun active-grid-operate (&optional pane operation)
  (multiple-value-bind (interface grid) (get-interface-active-grid pane)
    (cond ((operate-on-grid-p interface grid operation)
           ;(and (save-record-p grid) (active-grid-operate-p pane operation))
           (operate-on-grid interface grid operation))
          (grid
           (beep-impossible grid operation)))))

;;; Toolbar

(defparameter *record-commands-aplist*
 '((:new	:image 4  :title #L"New record"
		:mnemonic 0
		:accelerator #\Control-+
                :data #'new-record)
   (:insert	:image 18 :title #L"New before current"
		:mnemonic #-Russian 4 #+Russian 12
		:help-key #L"Insert a new blank record just before the current.")
   (:append	:image 16 :title #L"Paste append"
		:mnemonic #-Russian 0 #+Russian 12  
		:accelerator #\Meta-Insert
		:help-key #L"Create new records from the contents of the system clipboard.")
) )

(defun filter-tc-enabled (face component &optional grid)
 ;;; Called from update-toolbar for a component with :interaction = :multiple-selection
  ;; Value: T as the whole component is enabled
  ;; NB: (capi:search-for-item component :toggle-filter)  !
  (when (typep (orf grid (grid face)) 'sorted-grid)
    (let ((filter-items (filter-items grid)))
      ;(component (slot-value face 'record-filter-tc))
      ;; Press the :toggle-filter button
      (setf (capi:choice-selection component) (if (filter-state grid) '(1) ()))
      (capi:map-collection-items component
            (lambda (item) (unless (eq (capi:item-data item) :filter-by)
                           ;(case (capi:item-data item)
                           ;((:toggle-filter :destroy-filter)
                             (setf (capi:simple-pane-enabled item) filter-items))))
      t)))

(defun make-records-toolbar (&key grid divider-p enabled-function
                                  (filter t) (sort t) (find t)
                                  (new #5='(:new :append))
                                  (delete t) (revert t)
                                  refresh after)
 ;;; Args: enabled-function
  ;;		 For the entire toolbar, not for individual items or buttons.
  ;;       new   List of keys or sinle key, e.g. :new = only New button displayed,
  ;;		 T = (:new :append)
  ;;       after List of additional instances of toolbar-object placed at the right side.
  (flet ((image (item)
           (getf (cdr-assoq (first-or-self item) *record-commands-aplist*) :image))
         (data (item &aux (self (first-or-self item))
                          (default (getf (cdr-assoq self *record-commands-aplist*)
                                         :data self)))
           (if (consp item) (getf (cdr item) :data default) default))
         (tooltip (item &aux (default (getf (cdr-assoq (first-or-self item)
                                                       *record-commands-aplist*)
                                             :title)))
           (if (consp item) (getf (cdr item) :title default) default))
        )
   (let* ((callback (lambda (interface operation)
                      (operate-on-grid interface grid operation)))
          (record-filter-tc
           (when filter
             (make-instance 'capi:toolbar-component		;:name :record-filter-tc
              :items `(,(make-instance 'capi:toolbar-button  :data :filter-by
                                       :image 11  :tooltip #L"Filter by selection")
                       ,(make-instance 'capi:toolbar-button  :data :toggle-filter
                                       :image 12  :tooltip #L"Toggle filter on/off"
                                       :retract-callback callback)
                       ,(make-instance 'capi:toolbar-button  :data :destroy-filter
                                       :image 13  :tooltip #L"Destroy filter"))
              :interaction :multiple-selection))))
    (when filter
      (setf (capi:toolbar-object-enabled-function record-filter-tc)
             (lambda (interface) (filter-tc-enabled interface record-filter-tc grid))))
    (list (make-instance 'capi:toolbar  :name 'records-toolbar  :items
           `(,@(when sort
                 `(,(make-instance 'capi:toolbar-component	;:name 'record-sort-tc
			:items '(:sort-asc :sort-desc)
			:images '(9 10)
			:tooltips '(#L"Sort ascending" #L"Sort descending"))))
             ,@(when filter
                 `(,record-filter-tc))
             ,@(when find
                 `(,(make-instance 'capi:toolbar-component	;:name 'record-search-tc
			:items (list 'searcher)
			:images '(14)
			:tooltips '(#L"Find in table"))))
             ,(make-instance 'capi:toolbar-component	  	;:name 'record-operate-tc
			:items `(,#'save-record
                                 ,@(when revert (list #'revert-record))
                                 ,@(when delete (list #'delete-record-or-rows)))
			:images `(7
                                  ,@(when revert '(8))
                                  ,@(when delete '(5)))
			:tooltips `(#L"Save record"
                                    ,@(when revert '(#L"Revert record"))
                                    ,@(when delete '(#L"Delete record"))))
             ,@(when new
                 `(,(make-instance 'capi:toolbar-component	;:name 'record-new-tc
			:items (mapcar (lambda (item)
                                         (make-instance 'capi:toolbar-button
                                          :data (data item)
                                          :image (image item)
                                          :tooltip (tooltip item)))
                                       (if (eq new t) #5# (listify new))))))
             ,@(when refresh
                 `(,(ywi:doc-refresh-tc)))
             ,@(listify after))
           :selection-callback callback  :callback-type :interface-data
           :enabled-function enabled-function
           :dividerp divider-p
           :default-image-set *record-image-set*
           :button-width 18 :button-height 18)
          :title #L"Records"))))

#|
(defun update-records-toolbar (interface &optional (parent (slot-value interface
                                                                       'top-docking)))
  (when-let (toolbar (ywi:find-pane-by-name parent 'records-toolbar))
    (capi:update-toolbar toolbar)))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  RECORDS-MENU  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod operate-on-grid (interface (grid basic-grid) (operation (eql 'goto-record)))
  (if-bind (navigation-toolbar (slot-value grid 'navigation-toolbar))
    (let ((record-number-pane (record-number-pane navigation-toolbar)))
      (capi:set-pane-focus record-number-pane)
      (record-number-pane-on-edit record-number-pane))	; Lispworks5 Q: For programmatic?
    (when-let (record-number (capi:prompt-for-integer #L"Enter record number:"
                               :min 1 :max (nrows grid)
                               :popup-args `(:owner ,interface ,@ywi:+popup-args+)))
      (record-number-callback grid nil record-number))))

(defun filter-mi-enabled (face &optional grid)
  (and (typep (orf grid (grid face)) 'sorted-grid)
       (filter-items grid)))

(defun records-menu (interface &key grid enabled-function
                                    (filter t) (sort t)
                                    (new #5='(:new :append))
                                    (delete t)
                                    (revert t)
                                    refresh)
 ;;; Args: grid	 If specified, the menu gets bound to this grid;
  ;;		 otherwise works on (grid interface) computed dynamically.
  ;;       enabled-function
  ;;		 For the entire menu, not for individual items or buttons
  (let ((callback (lambda (face operation) (operate-on-grid face grid operation)))
        (callback-type :interface-data))
   (flet ((data (item &aux (self (first-or-self item))
                           (default (getf (cdr-assoq self *record-commands-aplist*)
                                          :data self)))
            (if (consp item) (getf (cdr item) :data default) default))
          (title (item &aux (default (getf (cdr-assoq (first-or-self item)
                                                        *record-commands-aplist*)
                                           :title)))
            (if (consp item) (getf (cdr item) :title default) default))
          (mnemonic (item &aux (default (getf (cdr-assoq (first-or-self item)
                                                        *record-commands-aplist*)
                                              :mnemonic 0)))
            (if (consp item) (getf (cdr item) :mnemonic default) default))
          (accelerator (item &aux (default (getf (cdr-assoq (first-or-self item)
                                                            *record-commands-aplist*)
                                                 :accelerator)))
            (if (consp item) (getf (cdr item) :accelerator default) default))
          (help-key (item &aux (default (getf (cdr-assoq (first-or-self item)
                                                         *record-commands-aplist*)
                                              :help-key)))
            (if (consp item) (getf (cdr item) :help-key default) default))
         )
     (make-instance 'capi:menu
	:title #L"Records"  :mnemonic 0
	:name 'records-menu
	:items
	`(,(make-instance 'capi:menu-component :items
            `(,(make-instance 'capi:menu-item :title #L"Save record"
		:mnemonic 0 :accelerator #\Shift-Return
		:data #'save-record
		:help-key nil)
              ,@(when revert
                  `(,(make-instance 'capi:menu-item 
			:title (load-time-value (string-append #L"Revert record" "..."))
			:mnemonic 0
			:data #'revert-record
			:help-key #L"Reset all the record cells to pre-edit values.")))
              ,@(when delete
                  `(,(make-instance 'capi:menu-item
			:title (load-time-value (string-append #L"Delete record" "..."))
			:mnemonic 0 :accelerator #\Control-\-
			:data #'delete-record-or-rows
        	        :help-key #L"If there are selected records, delete them; otherwise delete the current record."))))
            :enabled-function enabled-function)
          ,@(when new
              `(,(make-instance 'capi:menu-component
		 :items (mapcar (lambda (item)
                                  (make-instance 'capi:menu-item
                                        :data (data item)
                                        :title (title item)
                                        :mnemonic (mnemonic item)
                                        :accelerator (accelerator item)
                                        ;:enabled-function (enabled-function item)
					;:callback (callback item)
                                        :help-key (help-key item)))
                                (if (eq new t) #5# (listify new)))
                 :enabled-function enabled-function)))
          ,@(when sort
              `(,(make-instance 'capi:menu-component
		  :items
		  `(,(make-instance 'capi:menu-item :title #L"Sort ascending"
			:mnemonic #-Russian 5 #+Russian 19
			:data :sort-asc  :help-key nil)
		    ,(make-instance 'capi:menu-item :title #L"Sort descending"
			:mnemonic #-Russian 5 #+Russian 22
			:data :sort-desc  :help-key nil))
                  :enabled-function enabled-function)))
          ,@(when filter
              (let ((enabled (if grid
                                 (lambda (face) (filter-mi-enabled face grid))
                                 'filter-mi-enabled)))
                `(,(make-instance 'capi:menu-component	;:name 'record-filter-mc
		    :items
                    `(,(make-instance 'capi:menu-item :title #L"Filter by selection"
			:mnemonic 0 :data :filter-by ;:enabled-function enabled-function
			:help-key #L"Filter by selection (in accumulative manner).")
                      ,(make-instance 'capi:menu-item :title #L"Toggle filter on/off"
			:mnemonic #-Russian 0 #+Russian 2
			:data :toggle-filter
			:enabled-function enabled
			:retract-callback callback
			:help-key nil)
        	      ,(make-instance 'capi:menu-item :title #L"Destroy filter"
			:mnemonic 3 :data :destroy-filter
			:enabled-function enabled
			:help-key nil))
                    :interaction :multiple-selection
                    :selection-function (lambda (face &aux (grid (grid face)))
                                          ;; tick off the :toggle-filter item
                                          (and (typep grid 'sorted-grid)
                                               (filter-state grid)
                                               '(1)))
                    :enabled-function enabled-function))))
          ,(make-instance 'capi:menu-component		;:name 'record-select-mc
	    :items
            `(,(make-instance 'capi:menu-item :title #L"Select current record"
		:mnemonic 1 :data :select-record
		:help-key #+Russian "  ."
			  #-Russian nil)
              ,(make-instance 'capi:menu-item :title #L"Select all records"
		:mnemonic 11 :data :select-all-records
		:help-key #+Russian "   ."
			  #-Russian nil))
            :enabled-function enabled-function)
          ,(make-instance 'capi:menu-component 		;:name 'record-search-mc
            :items
            `(,(make-instance 'capi:menu-item
		:title (load-time-value (string-append #L"Find in table" "..."))
		:mnemonic 0 :data 'searcher :accelerator #\Control-\f
		:help-key nil) ;#L"Find in table or formulary"
              ,(make-instance 'capi:menu-item :title #L"Go to record..."
		:mnemonic 0 :data 'goto-record :accelerator #\Control-\g
		:help-key #+Russian "    ,   ,     ."
                          #-Russian nil)
             )
            :enabled-function enabled-function)
          ,@(when refresh
              `(,(ywi:doc-refresh-mi)))
          )
	:selection-callback callback
	:callback-type callback-type
	:menu-interface interface
	:help-key '(:ygrid . #L"Commands and operations on grid elements.")))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  EDIT-TOOLBAR, -MENU  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Methods invoked when the focus is outside an editing pane, e.g. 
;;; from text-display-column cell or when row or column selection is active.

;;; Cut

(defmethod capi:pane-interface-cut-p ((focus grid-component) (interface capi:interface))
  (operate-on-grid-p interface (capi:element-interface focus) :cut))
(defmethod capi:pane-interface-cut-object ((focus grid-component)
                                           (interface capi:interface))
  (operate-on-grid interface (capi:element-interface focus) :cut))

(defmethod operate-on-cell-p ((grid ygrid) record column (operation (eql :cut)))
  nil)
(defmethod operate-on-cell-p :around ((grid ygrid) (record record) (column column)
                                      (operation (eql :cut)))
  (and (allow-edits grid)
       (not (memq (enabled column) '(nil :read-only)))
       (column-nullable column)
       (cell-value grid record column)
       (call-next-method)))

(defmethod operate-on-cell :around ((grid ygrid) record column (operation (eql :cut)))
  (if (operate-on-cell-p grid record column operation)
      (call-next-method)
      (beep-impossible grid operation :name (ywi:object-title column))))

(defmethod operate-on-cell ((grid ygrid) (record record) (column column)
                            (operation (eql :cut)))
 ;;; DI 2011-Apr-15: Any substantial change should go through save-cell.
  (operate-on-cell grid record column :copy)
  (save-cell grid record column :value nil))
  ;(set-cell-value grid record column nil))

;;; Copy: generic cell copy works for many columns classes, except check-column

(defmethod capi:pane-interface-copy-p ((focus grid-component) (interface capi:interface))
  (operate-on-grid-p interface (capi:element-interface focus) :copy))

(defmethod capi:pane-interface-copy-object ((focus grid-component)
                                            (interface capi:interface))
  (operate-on-grid interface (capi:element-interface focus) :copy))

(defmethod operate-on-cell-p ((grid ygrid) (record record) (column column)
                              (operation (eql :copy)))
  (and (enabled column) (cell-value grid record column)))

(defmethod operate-on-cell ((grid ygrid) (record record) (column column)
                            (operation (eql :copy)))
  (multiple-value-bind (string plist) (cell-clip grid record column)
    (if (or string plist)
        (let ((value (cell-value grid record column)))
          (capi:set-clipboard (capi:top-level-interface grid) value string plist)
          value)
        (beep-impossible grid operation :name (ywi:object-title column)))))

;;; Paste

(defmethod capi:pane-interface-paste-p ((focus grid-component) (interface capi:interface))
  (operate-on-grid-p interface (capi:element-interface focus) :paste))

(defmethod capi:pane-interface-paste-object ((focus grid-component)
                                             (interface capi:interface))
  (operate-on-grid interface (capi:element-interface focus) :paste))

(defmethod operate-on-cell-p ((grid ygrid) record column (operation (eql :paste)))
  nil) ;(not (and (capi:clipboard-empty grid :string) (capi:clipboard-empty grid :value)))
(defmethod operate-on-cell-p :around ((grid ygrid) (record record) (column column)
                                      (operation (eql :paste)))
  (and (allow-edits grid) (not (memq (enabled column) '(nil :read-only)))
       (call-next-method)))

(defmethod operate-on-cell ((grid ygrid) (record record) (column column)
                            (operation (eql :paste)))
  (or (loop with interface = (capi:top-level-interface grid)
            for format in (cell-clip-formats column)
            thereis (and (not (capi:clipboard-empty grid format))
                         (set-cell-clip grid record column
                                        format (capi:clipboard interface format))))
      (beep-impossible grid operation :name (ywi:object-title column))))
 #|(if (capi:clipboard-empty interface :value)
      (dolist (format (cell-clip-formats column)
                      (beep-impossible grid operation :name (ywi:object-title column)))
        (when (and (not (capi:clipboard-empty grid format))
                   (set-cell-clip grid record column format (capi:clipboard grid format)))
          (return)))
      (multiple-value-bind (value condition) (clipboard-value-bad grid column record)
        (if condition
            (beep-impossible grid operation
                             :name (ywi:object-title column) :condition condition)
            (save-cell grid record column :value value))))	;:redraw-now|#

#|(defgeneric clipboard-value-bad (interface displayer &optional subscript)
 ;;; Filter the current Lisp value kept in the clipboard.
  ;; Values: 1) Converted value 
  ;;         2) False if fits fine, or generalized true if does not (e.g. condition?)
 (:method (interface displayer &optional subscript)
  (capi:clipboard interface :value)))|#

;;; Undo

(defmethod capi:pane-interface-undo-p ((focus grid-component) (interface capi:interface))
  (operate-on-grid-p interface (capi:element-interface focus) :undo))
(defmethod capi:pane-interface-undo ((focus grid-component) (interface capi:interface))
  (operate-on-grid interface (capi:element-interface focus) :undo))

(defmethod operate-on-grid-p (face (grid ygrid) (operation (eql :undo)))
  (and (not (or (selected-rows grid) (selected-columns grid)))
       (operate-on-cell-p grid (current-record grid) (current-column grid) operation)))

(defmethod operate-on-grid (face (grid ygrid) (operation (eql :undo)))
  (if (or (selected-rows grid) (selected-columns grid))
      (beep-impossible grid :undo)
      (undo-cell grid (current-record grid) (current-column grid))))

;;; Select All

(defmethod operate-on-cell-p ((grid ygrid) record column (operation (eql :select-all)))
  nil)

(defmethod operate-on-cell-p ((grid ygrid) (record record) (column text-input-column-mixin)
                              (operation (eql :select-all)))
  t)

(defmethod operate-on-cell ((grid ygrid) (record record) (column text-input-column-mixin)
                            (operation (eql :select-all)))
  (let* ((pinboard (pinboard grid))
         (overlay (overlay pinboard)))
    (if (and overlay					; cell-current-p?
             (eq (current-displayer pinboard) column)
             (eq (current-subscript pinboard) record))
        (capi:pane-interface-select-all overlay (capi:top-level-interface grid))
        ;(capi:call-editor overlay "Mark Whole Buffer")
        (ywi:beep-impossible grid :select-all :name (ywi:object-title column)))))

