;;;-*- Mode: Lisp; -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Filter and sort facility
;;;
;;; RULES OF THUMB:
;;; 1. Current coulumn is ratained.
;;; 2. These functions must not switch input focus except when :focus t argument is passed.
;;;
;;; CAPI NB: We inherit form capi:sorted-object and exploit capi::sort-state,
;;;          but other CAPI built-ins are not used.

(in-package :ystok.grid)

(defclass sorted-grid (ygrid capi:sorted-object)
 ((filter-state :accessor filter-state :initform nil)	; t=on, nil=off
  (filter-items :accessor filter-items :initform ())
  (records-off  :initform ())))		; list of records being off 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  FILTER  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; For every column, maximum one predicate (conjunct) allowed.
;;; When filter is being made stronger, the sort state is being retained;
;;; when weaker - eliminated.

(defstruct filter-item
  column
  predicate	; function: record-data -> boolean
  value)
  ;type		; :region or :full

(defun %filter-on (grid &optional current focus)
 ;;; Args: current If true, try to invoke current-callback.
  ;;               (True is usally passed when the current record is off.)
  (with-slots (filter-state filter-items nrows current-row) grid
   (when (and filter-items (grid-confirm-destroy grid))
    (when-let (predicate (apply #'fand
                                (mapcar #'filter-item-predicate filter-items)))
      (deselect-section grid :rows nil)				; :redraw-now nil
      (let ((records (records grid))
            (records-off (slot-value grid 'records-off))
            (current-record (row-record grid))
            (auto-record (record-auto-p grid (1- nrows)))) 
        (loop with filter-pointer fixnum = 0
              for row fixnum upfrom 0 below (if auto-record (1- nrows) nrows)
              for record = (aref records row)
                  ;for datum = (call-data-reader grid (record-data record) column)
                  ;if (equal datum value)
              if (funcall predicate (record-data record))
              do (setf (aref records filter-pointer) record
                       (record-row-number record) filter-pointer)
                 (incf filter-pointer)
              else do (setf (record-row-number record) -1)
                      (push record records-off)
              finally
              (when auto-record
                (setf (aref records filter-pointer) auto-record
                      (record-row-number auto-record) filter-pointer)
                (incf filter-pointer))
              (setf (fill-pointer records) filter-pointer
                    nrows filter-pointer))
        (setf (slot-value grid 'records-off) records-off
              current-row (if (or (null current-record)
                                  (eql (record-row-number current-record) -1))
                              0 
                              (record-row-number current-record))
              (current-subscript (pinboard grid)) (if (plusp nrows) (%current-record grid))
              filter-state t)
        (%invalidate-grid grid (if focus :focus t))
        (when current
          (%record-on-current grid current-record)))))))

(defun %filter-off (grid &optional current focus (redraw-now focus))
 ;;; Remove filter and show all records
  ;; Args: redraw-now False - usually on applying new filter just before %filter-off.
  (with-slots (filter-state nrows current-row capi::sort-state) grid
    (unless (and filter-state
                 (grid-confirm-destroy grid))
      (return-from %filter-off))
    (let ((records (records grid))
          (current-record (row-record grid))
          (auto-record (record-auto-p grid (1- nrows))))
      (when auto-record (decf (fill-pointer records)))
      ;; Is it much safer to pop records by one from the (slot-value grid 'records-off)?
      (dolist (record (nreverse (slot-value grid 'records-off)))
        (setf (record-row-number record) (fill-pointer records))
        (vector-push-extend record records))
      (when auto-record
        (setf (record-row-number auto-record) (fill-pointer records))
        (vector-push-extend auto-record records))
      (setf (slot-value grid 'records-off) ()
            nrows (fill-pointer records)
            current-row (if (and current-record (neq current-record auto-record))
                            (record-row-number current-record)
                            0)
            (current-subscript (pinboard grid)) (%current-record grid)
            filter-state nil
            capi::sort-state (capi::make-sort-state))	; eliminate sort state
      (when redraw-now
        (%invalidate-grid grid (if focus :focus t))
        (when current
          (%record-on-current grid current-record))))))


(defmethod filter-records ((grid sorted-grid) (column basic-column) 
                           &key (state t) test test-not (value nil value-specified-p)
                                (current t) (focus t) stronger)
 ;;; Args: state    True = Add or change conjunct on the column,
  ;;		    False = remove old conjunct if any
  ;;       value-specified-p
  ;;		    True = test or test-not is a function of two parameters
  ;; 		    False= test or test-not is a function of one parameter.
  ;;       stronger Optimization hint: true = the caller guarantees that condition
  ;;		    has made stronger and there is not need to %filter-off before -on.
  (with-slots (filter-items) grid
   (let ((filter-item (find column filter-items :key #'filter-item-column))
          ;(data-reader (column-data-reader column))	; analyse data-reader types?
        )
    (cond (state			; filter on
           (let ((predicate
                  (cond (test
                           (if value-specified-p
                               (lambda (data)
                                 (funcall test (call-data-reader grid data column) value))
                               (lambda (data)
                                 (funcall test (call-data-reader grid data column)))))
                        (test-not
                         (if value-specified-p
                             (lambda (data)
                               (not (funcall test-not
                                             (call-data-reader grid data column) value)))
                             (lambda (data)
                               (not (funcall test-not
                                             (call-data-reader grid data column))))))
                        (value-specified-p
                         (lambda (data)
                           (equal (call-data-reader grid data column) value)))
                        (t					; avoid NIL as predicate
                         (error "Neither test(-not) nor value specified for filter-records")))))
             (cond (filter-item					; reuse the item
                    (setf (filter-item-value filter-item) value
                          (filter-item-predicate filter-item) predicate)
                    (unless stronger
                      (capi:apply-in-pane-process grid #'%filter-off grid)))
                   (t
                    (push (make-filter-item :column column
                                            :predicate predicate
                                            :value value)
                          filter-items)))
             (capi:apply-in-pane-process grid #'%filter-on grid current focus)))
          (filter-item   		; filter off - remove item or show all records
           (cond ((setf filter-items (delete filter-item filter-items))
                  (capi:apply-in-pane-process grid #'%filter-off grid)
                  (capi:apply-in-pane-process grid #'%filter-on grid current focus))
                 (t
                  (capi:apply-in-pane-process grid #'%filter-off grid current focus t))))
))) )  

(defmethod filter-records ((grid sorted-grid) (column null)
                           &key (state t) test test-not (value nil value-specified-p)
                                (current t) (focus t) stronger)
 ;;; Provide predicate on the data as a whole instead of some column.
  ;; Args: value-specified-p True = test or test-not is a function of two parameters
  ;; 			     False= test or test-not is a function of one parameter.
  (with-slots (filter-items) grid
    (let ((filter-item (find column filter-items :key #'filter-item-column))
          ;(data-reader (column-data-reader column))	; analyse data-reader types?
         )
      (if state		;; Filter on
          (when-let (predicate (cond (test
                                      (if value-specified-p
                                          (lambda (data) (funcall test data value))
                                          test))
                                     (test-not
                                      (if value-specified-p
                                          (lambda (data)
                                            (not (funcall test-not data value)))
                                          (complement test-not)))))
                       ;(t (lambda (data) (equal data value))))))
            (cond (filter-item					; reuse the item
                   (unless stronger
                     (capi:apply-in-pane-process grid #'%filter-off grid))
                   (setf (filter-item-value filter-item) value
                         (filter-item-predicate filter-item) predicate))
                  (t (push (make-filter-item :column column
                                             :predicate predicate
                                             :value value)
                           filter-items)))
            (capi:apply-in-pane-process grid #'%filter-on grid current focus))
          ;; else - Filter off - show all records
          (when filter-item
            (cond ((setf filter-items (delete filter-item filter-items))
                   (capi:apply-in-pane-process grid #'%filter-off grid)
                   (capi:apply-in-pane-process grid #'%filter-on grid current focus))
                  (t
                   (capi:apply-in-pane-process grid #'%filter-off grid current focus t)))))
) ) )


;(defmethod filter-records ((grid sorted-grid) (column (eql :all))
;                           &key (state :off) redraw-now)
  ;(unless (capi:apply-in-pane-process grid #'grid-confirm-destroy grid)
  ;  (return-from filter-records))

(defgeneric filter-by-selection (grid column)
 ;;; By default, compare by equal to the current cell value.
 (:method ((grid sorted-grid) (column column))
  (filter-records grid column :state t
                  :value (cell-value grid (current-row grid) column)))
 (:method :around ((grid sorted-grid) column)
  (declare (ignore column))
  (if (< (current-row grid) (nrows grid))
      (call-next-method)
      (beep))) )

(defmethod filter-by-selection ((grid sorted-grid) (column text-column))
  (when-let (overlay (overlay (pinboard grid)))
    (if-bind (region (editor:buffer-region-as-string (capi:editor-pane-buffer overlay)))
      (filter-records grid column :state t
                      :test (lambda (subject pattern)
                              (when (stringp subject) (search pattern subject)))
                      :value region)
      (let ((value (cell-value grid (current-row grid) column)))
        (if-bind (column-test-function (column-test-function column))
          (if (column-nullable column)
              (filter-records grid column :state t
                              :test (if value
                                        (lambda (arg)
                                          (when arg
                                            (funcall column-test-function arg value)))
                                        #'null))
              (filter-records grid column :state t :value value
                              :test column-test-function))
          (filter-records grid column :state t :value value))))))

(defmethod filter-by-selection ((grid sorted-grid) (column check-column))
  (let* ((value (cell-value grid (current-row grid) column))
         (test (if (column-nullable column)
                   (case value
                     ((nil) #'null)
                     ((:false 0) (lambda (datum) (member datum #1='(:false 0))))
                     (otherwise (lambda (datum) (and datum (not (member datum #1#))))))
                   (case value
                     ((nil :false 0) (lambda (datum) (member datum #2='(nil :false 0))))
                     (otherwise (lambda (datum) (not (member datum #2#))))))))
      (filter-records grid column :state t :test test)))

(defmethod toggle-filter ((grid sorted-grid) &optional (state (not (filter-state grid)))
                                                       (current t) (focus t))
  (if state
      (capi:apply-in-pane-process grid #'%filter-on grid current focus)
      (capi:apply-in-pane-process grid #'%filter-off grid current focus t)))

(defun destroy-filter (grid &key (redraw-now t) (current t) (focus t))
  (setf (slot-value grid 'filter-items) nil)
  (capi:apply-in-pane-process grid #'%filter-off grid current focus redraw-now))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  SORT  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Sorting is performed according to cell data, not printing representation
;;; If you need sorting according text, compose column-print-function explicitely
;;; into your :sort-asc and :sort-desc.

;;; Numbers: nil < number
(defun nullable< (x y)
  (cond ((null x) (not (null y)))
        ((null y) nil)
        (t (< x y))))

(defun nullable> (x y) 
  (cond ((null y) (not (null x)))
        ((null x) nil)
        (t (> x y))))

;;; Logical: nil < :false < :true
(defun logical< (x y)
  (case x ((nil) y)
          ((0 :false) (case y ((0 :false nil) nil) (otherwise t)))))
                  
(defun logical> (x y)
  (case y ((nil) x)
          ((0 :false) (case x ((0 :false nil) nil) (otherwise t)))))

;;; Characters: nil < char
(defun nullable-char< (x y)
  (cond ((null x) (not (null y)))
        ((null y) nil)
        (t (char< x y))))

(defun nullable-char> (x y) 
  (cond ((null y) (not (null x)))
        ((null x) nil)
        (t (char> x y))))

;;; The following are used for "system" (ASCII-only) strings as well as symbols.
;;; lw:locale-... do not work for symbols!
(defun nullable-string-lessp (x y)
  (cond ((null x) (not (null y)))
        ((null y) nil)
        (t (string-lessp x y))))

(defun nullable-string-greaterp (x y) 
  (cond ((null y) (not (null x)))
        ((null x) nil)
        (t (string-greaterp x y))))

(defgeneric sort-records (grid operation sorting-description &key focus))

(defmethod sort-records :around (grid operation sorting-description &key focus)
  (declare (ignorable focus))
  (when (grid-confirm-destroy grid)
    (deselect-section grid :rows nil)			; :redraw-now nil
    (call-next-method)))

(defmethod sort-records ((grid sorted-grid) (operation symbol) (sorting-description null)
                         &key (focus t))
  (declare (ignore sorting-description))
  (when-let (column (current-column grid))
    (sort-records grid operation column :focus focus)))

(defmethod sort-records ((grid sorted-grid) (operation capi::sort-state)
                         sorting-description &key (focus t))
  (capi:apply-in-pane-process grid #'%sort-records grid operation nil focus))

(defmethod sort-records ((grid sorted-grid) (operation symbol) (column column)
                         &key (focus t))
  (with-slots (capi::sort-descriptions) grid
    (with-slots (sort-asc sort-desc) column
      (if (or sort-asc sort-desc)
          (let ((sorting-description (first (member column capi::sort-descriptions
                                              :key #'capi::sorting-description-type))))
            (unless sorting-description
	      ;; Compile key function by closing lexically over data-reader
              (push (setq sorting-description
                       (let ((data-reader (column-data-reader column)))
                          (capi:make-sorting-description
                           :type column
                           :key (if (integerp data-reader)
                                    (lambda (record)
                                      (elt (record-data record) data-reader))
                                    (lambda (record)
                                      (let ((*record-data* (record-data record))
                                            (*column* column))
                                        (funcall data-reader *record-data*))))
                           :sort sort-asc
                           :reverse-sort sort-desc)))
                    capi::sort-descriptions))
            (capi:apply-in-pane-process grid #'%sort-records grid
                 (case operation
                   ((:asc :sort-asc)
                    (capi::make-sort-state :type column))
                   ((:desc :sort-desc)
                    (capi::make-sort-state :type column :reversep t))
                   (otherwise
                    (capi::make-sort-state)))
                 sorting-description focus))
          (progn
            (message-area-format grid
                                 #L"Sorting predicate is not defined for column ~s."
                                 (column-title column))
            (beep))))))

(defmethod sort-records ((grid sorted-grid) (operation symbol) (column-name symbol)
                         &key (focus t))
  (sort-records grid operation (find-column grid column-name) :focus focus))

(defmethod sort-records ((grid sorted-grid) (operation symbol) (column-list cons)
                         &key (focus t))
  (declare (ignore focus))
  (beep-not-implemented grid "Multi-column sort"))

(defun reverse-sorting (grid &optional (column (%current-column grid)))
 ;;; If we are sorting by the column given now, reverse existing sorting; 
  ;; otherwise sort ascending if possible
  ;; Used by: e.g. as header-click-callback
  (with-slots (capi::sort-state) grid
    (sort-records grid 
                  (if (and capi::sort-state
                           (eq (capi::sort-state-type capi::sort-state) column))
                      (if (capi::sort-state-reversep capi::sort-state) :asc :desc)
                      :asc)
                  column)))

(defun %sort-records (grid &optional (state (slot-value grid 'capi::sort-state))
                                     sorting-description
                                     focus)
  (with-slots (records capi::sort-descriptions capi::sort-state) grid
    (unless sorting-description
      (setq sorting-description (first (member (capi::sort-state-type state)
                                               capi::sort-descriptions
                                               :key #'capi::sorting-description-type))))
    (let ((auto-record (record-auto-p grid (1- (nrows grid))))
          (current-record (row-record grid))
          (*grid* grid))
      (when auto-record (decf (fill-pointer records)))
      (unwind-protect
          (if (eq (capi::sort-state-type state) :unsorted)
              nil  
              (setf records
               (funcall (capi::sorting-description-sort-function sorting-description)
                       records
                       (if (capi::sort-state-reversep state)
                           (capi::sorting-description-reverse-sort sorting-description)
                           (capi::sorting-description-sort sorting-description))
                       :key (capi::sorting-description-key sorting-description))))
        (loop for record across records				; renumber records
              and row fixnum upfrom 0
              do (setf (record-row-number record) row))
        (setf (current-row grid) (if current-record		; restore current row
                                     (record-row-number current-record)
                                     0)
              capi::sort-state state)				; save the state
        (when auto-record (incf (fill-pointer records))))	; restore auto-record
      (setf (current-subscript (pinboard grid)) current-record))
    (%invalidate-grid grid (if focus :focus t))))
    #|;; Assign new sorting markers
      (unless (symbolp (capi::sort-state-type state))
        (dolist (column (listify (capi::sort-state-type state)))
          (setf (column-sorted column) column-sorted)))|#

(defmethod draw-cell :after ((grid sorted-grid) (column-header column-header-row)
                             column port x
                             &optional graphics-args)
 ;;;  Draw sorting trianlge at bottom left or right corner if the column 
  ;;  corresponds to the sort-state type.
  ;; Args: x  Left visible coordinate of cell rectangle
  (declare (ignore graphics-args)
           #-debug (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0)))
  (when-let (state (capi::sorted-object-state grid))
    (when (eq (capi::sort-state-type state) column)
      (let ((column-width (column-width column))
            (height (- (gp:port-height column-header) 2))
            (x (the fixnum (or x (column-left grid column)))))
       (declare (fixnum column-width height x))
       (with-cell-mask (mask-left mask-width) grid column x column-width
         (when (= mask-width 0) (return-from draw-cell))
         (unless port (setq port column-header))
         (with-translated-mask (port nil :left mask-left :top 1
                                         :width mask-width :height height)
           (gp:draw-image port
                          (if (column-selected column)
                              (column-header-image-selected column-header)
                              (column-header-image column-header))
                          (case (column-header-align column)
                            (:right (1+ x))
                            (otherwise (the fixnum (+ x column-width -3
                                                      (- +column-header-image-width+)))))
                          (the fixnum (- height 1 +column-header-image-height+))
                          :from-x (if (capi::sort-state-reversep state)
                                      +column-header-image-width+
                                      0)
                          ;:to-height (- height 2)
                          :to-width +column-header-image-width+
                          :from-width +column-header-image-width+
                          :operation (if (column-selected column) boole-orc1 boole-1))
) ) ) ) ) )

(defmethod wipe-grid :before ((grid sorted-grid) &key sort &allow-other-keys)
 ;;; Args: sort - if true, does not drop current sort state
  (with-slots (filter-items filter-state records-off) grid
    (unless sort
      (setf (capi::sorted-object-state grid) (capi::make-sort-state)))
    (nilf records-off filter-state filter-items)))

(defmethod find-record ((grid sorted-grid) data &key (test #'eq) (filter :on))
  (or (and (or (eq filter :on) (eq filter :all))
           (find data (records grid) :key #'record-data :test test)) ; respect fill-pointer
      (and (or (eq filter :off) (eq filter :all))
           (find data (slot-value grid 'records-off) :key #'record-data :test test))))
